

import argparse
import numpy as np
import os
import cv2
from tqdm import tqdm
import json
import multiprocessing as mp

# Disable scientific notation
np.set_printoptions(suppress=True)

def intrinsics_to_camera(K):
    px, py = K[0, 2], K[1, 2]
    fx, fy = K[0, 0], K[1, 1]
    return {
        "model": "PINHOLE",
        "width": int(2 * px),
        "height": int(2 * py),
        "params": [fx, fy, px, py],
    }

def angle_error_vec(v1, v2):
    n = np.linalg.norm(v1) * np.linalg.norm(v2)
    return np.rad2deg(np.arccos(np.clip(np.dot(v1, v2) / n, -1.0, 1.0)))

def angle_error_mat(R1, R2):
    cos = (np.trace(np.dot(R1.T, R2)) - 1) / 2
    cos = np.clip(cos, -1., 1.)  # numercial errors can make it out of bounds
    return np.rad2deg(np.abs(np.arccos(cos)))

def compute_pose_error(T_0to1, R, t):
    R_gt = T_0to1[:3, :3]
    t_gt = T_0to1[:3, 3]
    error_t = angle_error_vec(t, t_gt)
    error_t = np.minimum(error_t, 180 - error_t)  # ambiguity of E estimation
    error_R = angle_error_mat(R, R_gt)
    return error_t, error_R

def estimate_pose(kpts0, kpts1, K0, K1, thresh, conf=0.99999, type='poselib'):
    if len(kpts0) < 5:
        return None
    if type == 'poselib':
        import poselib
        (pose,details) = poselib.estimate_relative_pose(
            kpts0.tolist(), 
            kpts1.tolist(),
            intrinsics_to_camera(K0),
            intrinsics_to_camera(K1),
            ransac_opt={
                'max_iterations': 10000, # default 100000
                'success_prob': conf, # default 0.99999
                'max_epipolar_error': thresh, # default 1.0
            },
            bundle_opt={  # all defaults
                },
            )
        ret = (pose.R, pose.t, details['inliers'])

    elif type == 'opencv':
        f_mean = np.mean([K0[0, 0], K1[1, 1], K0[0, 0], K1[1, 1]])
        norm_thresh = thresh / f_mean

        kpts0 = (kpts0 - K0[[0, 1], [2, 2]][None]) / K0[[0, 1], [0, 1]][None]
        kpts1 = (kpts1 - K1[[0, 1], [2, 2]][None]) / K1[[0, 1], [0, 1]][None]

        E, mask = cv2.findEssentialMat(
            kpts0, kpts1, np.eye(3), threshold=norm_thresh, prob=conf,
            method=cv2.RANSAC)

        assert E is not None

        best_num_inliers = 0
        ret = None
        for _E in np.split(E, len(E) / 3):
            n, R, t, _ = cv2.recoverPose(
                _E, kpts0, kpts1, np.eye(3), 1e9, mask=mask)
            if n > best_num_inliers:
                best_num_inliers = n
                ret = (R, t[:, 0], mask.ravel() > 0)
    else:
        raise NotImplementedError

    return ret

def estimate_pose_parallel(args):
    return estimate_pose(*args)

def pose_auc(errors, thresholds):
    sort_idx = np.argsort(errors)
    errors = np.array(errors.copy())[sort_idx]
    recall = (np.arange(len(errors)) + 1) / len(errors)
    errors = np.r_[0., errors]
    recall = np.r_[0., recall]
    aucs = []
    for t in thresholds:
        last_index = np.searchsorted(errors, t)
        r = np.r_[recall[:last_index], recall[last_index-1]]
        e = np.r_[errors[:last_index], t]
        aucs.append(np.trapz(r, x=e)/t)
    return aucs

def pose_accuracy(errors, thresholds):
    return [np.mean(errors < t) * 100 for t in thresholds]

def get_relative_transform(pose0, pose1):
    R0 = pose0[..., :3, :3] # Bx3x3
    t0 = pose0[..., :3, [3]] # Bx3x1

    R1 = pose1[..., :3, :3] # Bx3x3
    t1 = pose1[..., :3, [3]] # Bx3x1
    
    R_0to1 = R1.transpose(-1, -2) @ R0 # Bx3x3
    t_0to1 = R1.transpose(-1, -2) @ (t0 - t1) # Bx3x1
    T_0to1 = np.concatenate([R_0to1, t_0to1], axis=-1) # Bx3x4

    return T_0to1

class Scannet1500:
    default_config = {
        'scannet_path': os.path.abspath(os.path.join(os.path.dirname(__file__), '../../data/ScanNet/scannet_test_1500')),
        'gt_path': os.path.abspath(os.path.join(os.path.dirname(__file__), '../../data/ScanNet/test.npz')),
        'pose_estimator': 'poselib', # poselib, opencv
        'cache_images': True,
        'ransac_thresholds': [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0],
        'pose_thresholds': [5, 10, 20],
        'max_pairs': -1,
        'output': './output/scannet/',
        'n_workers': 8,
    }

    def __init__(self, config={}) -> None:
        self.config = {**self.default_config, **config}
        if not os.path.exists(self.config['scannet_path']):
            raise RuntimeError(
                f"Dataset {self.config['scannet_path']} does not exist! \n \
                  > If you didn't download the dataset, use the downloader tool: python3 -m modules.dataset.download -h")
        
        self.pairs = self.read_gt()

        os.makedirs(self.config['output'], exist_ok=True)

        if self.config['n_workers'] == -1:
            self.config['n_workers'] = mp.cpu_count()

        self.image_cache = {}
        if self.config['cache_images']:
            self.load_images()

    def load_images(self):
        for pair in tqdm(self.pairs, desc='Caching images'):
            if pair['image0'] not in self.image_cache:
                self.image_cache[pair['image0']] = cv2.imread(pair['image0'])
            if pair['image1'] not in self.image_cache:
                self.image_cache[pair['image1']] = cv2.imread(pair['image1'])

    def read_image(self, path):
        if self.config['cache_images']:
            return self.image_cache[path]
        else:
            return cv2.imread(path)

    def read_gt(self):
        pairs = []
        gt_poses = np.load(self.config['gt_path'])
        names = gt_poses['name']
        
        for i in range(len(names)):
                scene_id = names[i, 0]
                scene_idx = names[i, 1]
                scene = f'scene{scene_id:04d}_{scene_idx:02d}'
            
                image0 = str(int(names[i, 2]))
                image1 = str(int(names[i, 3]))
                
                K0 = np.loadtxt(
                    os.path.join(self.config['scannet_path'], 'scannet_test_1500', scene,  'intrinsic/intrinsic_color.txt')
                )
                K1 = K0

                pose_0 = np.loadtxt(
                    os.path.join(self.config['scannet_path'], 'scannet_test_1500', scene, 'pose', image0 + '.txt')
                )
                pose_1 = np.loadtxt(
                    os.path.join(self.config['scannet_path'], 'scannet_test_1500', scene, 'pose', image1 + '.txt')
                )
                T_0to1 = get_relative_transform(pose_0, pose_1)
                
                pairs.append({
                    'image0': os.path.join(self.config['scannet_path'], 'scannet_test_1500', scene, 'color', image0 + '.jpg'),
                    'image1': os.path.join(self.config['scannet_path'], 'scannet_test_1500', scene, 'color', image1 + '.jpg'),
                    'K0': K0,
                    'K1': K1,
                    'T_0to1': T_0to1,
                })

        return pairs

    def extract_and_save_matches(self, matcher_fn, name='', force=False):
        all_matches = []
        if name == '':
            name = matcher_fn.__name__

        fname = os.path.join(self.config['output'], f'{name}_matches.npz')
        if not force and os.path.exists(fname):
            return np.load(fname, allow_pickle=True)['all_matches']

        for pair in tqdm(self.pairs, desc='Extracting matches'):
            image0 = self.read_image(pair['image0'])
            image1 = self.read_image(pair['image1'])
            mkpts0, mkpts1 = matcher_fn(image0, image1)

            all_matches.append({
                'image0': pair['image0'],
                'image1': pair['image1'],
                'mkpts0': mkpts0,
                'mkpts1': mkpts1,
            })
        
        np.savez(fname, all_matches=all_matches)
        
        return all_matches

    def run_benchmark(self, matcher_fn, name='', force=False):
        if name == '':
            name = matcher_fn.__name__

        all_matches = self.extract_and_save_matches(matcher_fn, name=name, force=force)

        aucs_by_thresh = {}
        accs_by_thresh = {}
        for ransac_thresh in self.config['ransac_thresholds']:

            fname = os.path.join(self.config['output'], f'{name}_{self.config["pose_estimator"]}_{ransac_thresh}.txt')
            # check if exists and has the right number of lines
            if not force and os.path.exists(fname) and len(open(fname, 'r').readlines()) == len(self.pairs):
                errors = []
                with open(fname, 'r') as f:
                    lines = f.readlines()
                    for line in lines:
                        line = line.replace('\n', '')
                        err_t, err_R = line.split(' ')
                        errors.append([float(err_t), float(err_R)])
            # redo the benchmark
            else:
                errors = []
                pairs = self.pairs
                errors_file = open(fname, 'w')

                # do the benchmark in parallel
                if self.config['n_workers'] != 1:

                    pool = mp.Pool(self.config['n_workers'])
                    pool_args = [ (all_matches[pair_idx]['mkpts0'], all_matches[pair_idx]['mkpts1'], pair['K0'], pair['K1'], ransac_thresh) for pair_idx, pair in enumerate(pairs) ]
                    results = list(tqdm(pool.imap(estimate_pose_parallel, pool_args), total=len(pool_args), desc=f'Running benchmark for th={ransac_thresh}', leave=False))
                    pool.close()
                    

                    for pair_idx, ret in enumerate(results):
                        if ret is None:
                            err_t, err_R = np.inf, np.inf
                        else:
                            R, t, inliers = ret
                            pair = pairs[pair_idx]
                            err_t, err_R = compute_pose_error(pair['T_0to1'], R, t)
                        errors_file.write(f'{err_t} {err_R}\n')
                        errors.append([err_t, err_R])
                # do the benchmark in serial
                else:
                    for pair_idx, pair in tqdm(enumerate(pairs), desc=f'Running benchmark for th={ransac_thresh}', leave=False, total=len(pairs)):
                        mkpts0 = all_matches[pair_idx]['mkpts0']
                        mkpts1 = all_matches[pair_idx]['mkpts1']
                        ret = estimate_pose(mkpts0, mkpts1, pair['K0'], pair['K1'], ransac_thresh)
                        if ret is None:
                            err_t, err_R = np.inf, np.inf
                        else:
                            R, t, inliers = ret
                            err_t, err_R = compute_pose_error(pair['T_0to1'], R, t)
                        errors_file.write(f'{err_t} {err_R}\n')
                        errors_file.flush()
                        errors.append([err_t, err_R])

                errors_file.close()

            # compute AUCs
            errors = np.array(errors)
            errors = errors.max(axis=1) 
            aucs = pose_auc(errors, self.config['pose_thresholds'])
            accs = pose_accuracy(errors, self.config['pose_thresholds'])
            aucs = {k: v*100 for k, v in zip(self.config['pose_thresholds'], aucs)}
            accs = {k: v for k, v in zip(self.config['pose_thresholds'], accs)}
            aucs_by_thresh[ransac_thresh] = aucs
            accs_by_thresh[ransac_thresh] = accs

            # dump summary for this method
            summary = {
                'name': name,
                'aucs_by_thresh': aucs_by_thresh,
                'accs_by_thresh': accs_by_thresh,
            }
            json.dump(summary, open(os.path.join(self.config['output'], f'{name}_{self.config["pose_estimator"]}_summary.json'), 'w'), indent=2)

        return aucs_by_thresh

def get_xfeat():
    from modules.xfeat import XFeat
    xfeat = XFeat()
    return xfeat.match_xfeat

def get_xfeat_star():
    from modules.xfeat import XFeat
    xfeat = XFeat(top_k=10_000)
    return xfeat.match_xfeat_star

def get_alike():
    from third_party import alike_wrapper as alike
    return alike.match_alike

def print_fancy(d):
    print(json.dumps(d, indent=2))

def parse():
    parser = argparse.ArgumentParser()
    
    parser.add_argument("--scannet_path", type=str, required=True, help="Path to the Scannet 1500 dataset")
    parser.add_argument("--output", type=str, default="./output/scannet/", help="Path to the output directory")
    parser.add_argument("--max_pairs", type=int, default=-1, help="Maximum number of pairs to run the benchmark on")
    parser.add_argument("--force", action='store_true', help="Force running the benchmark again")
    parser.add_argument("--pose_estimator", type=str, default='poselib', help="Which pose estimator to use: poselib, opencv", choices=['poselib', 'opencv'])
    
    parser.add_argument("--show", action='store_true', help="Show the matches")
    parser.add_argument("--accuracy", action='store_true', help="Show the accuracy instead of AUC")
    parser.add_argument("--filter", type=str, nargs='+', help="Filter the results by the given names")
    
    return parser.parse_args()

if __name__ == "__main__":
    args = parse()

    if not args.show:
        scannet = Scannet1500({
            'scannet_path': args.scannet_path,
            'gt_path': args.scannet_path + "/test.npz",
            'cache_images': False,
            'output': args.output,
            'max_pairs': args.max_pairs,
            'pose_estimator': args.pose_estimator,
            'ransac_thresholds': [0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0],
            'n_workers': 8,
        })

        functions = {
            'xfeat': get_xfeat(),
            'xfeat_star': get_xfeat_star(),
            'alike': get_alike(),
        }

        # save all results to a file
        all_results = {}

        for name, fn in functions.items():
            print(name)
            result = scannet.run_benchmark(matcher_fn = fn, name=name, force=args.force)
            all_results[name] = result

        json.dump(all_results, open(os.path.join(args.output, 'summary.json'), 'w'), indent=2)

    if args.show:
        import glob
        import pandas as pd
        dataset_name = 'scannet'
        all_summary_files = glob.glob(os.path.join(args.output, "**_summary.json"), recursive=True)

        if args.filter:
            all_summary_files = [f for f in all_summary_files if any([fil in f for fil in args.filter])]

        dfs = []
        names = []
        estimators = []
        metric_key = 'aucs_by_thresh'
        if args.accuracy:
            metric_key = 'accuracies_by_thresh'    
        for summary in all_summary_files:
            summary_data = json.load(open(summary, 'r'))
            if metric_key not in summary_data:
                continue
            aucs_by_thresh = summary_data[metric_key]

            estimator = 'poselib'
            if 'opencv' in summary:
                estimator = 'opencv'

            #make sure everything is float
            for thresh in aucs_by_thresh:
                for k in aucs_by_thresh[thresh]:
                    if isinstance(aucs_by_thresh[thresh][k], str):
                        aucs_by_thresh[thresh][k] = float(aucs_by_thresh[thresh][k].replace(' ', ''))

            # find best threshold based on the 5, 10, 20 mAP and everything is float
            df = pd.DataFrame(aucs_by_thresh).T.astype(float)
            df['mean'] = df.mean(axis=1)
            # create a string column called estimator
            cols = df.columns.tolist()
            dfs.append(df)
            names.append(summary_data['name'])
            estimators.append(estimator)

        # use each col as the main col to determine the best threshold
        # for col in cols:
        col = 'mean'

        final_df = pd.DataFrame()
        # add cols
        final_df['name'] = names
        final_df['best_thresh'] = ''
        final_df['estimator'] = estimators
        final_df[cols] = -1.0

        for df, name, estimator in zip(dfs, names, estimators):
            best_thresh = df[col].idxmax()
            best_results = df.loc[best_thresh]

            # now update the best_thresh based on the estimator
            final_df.loc[(final_df['name'] == name) & (final_df['estimator'] == estimator), 'best_thresh'] = best_thresh
            for _col in cols:
                final_df.loc[(final_df['name'] == name) & (final_df['estimator'] == estimator), _col] = best_results[_col]

        # sort by mean
        final_df = final_df.sort_values(by=['mean'])
        # reset index
        final_df = final_df.reset_index(drop=True)

        # drop estimator column
        final_df = final_df.drop(columns=['estimator'])

        # set max float precision to 1
        final_df = final_df.round(1)

        print(f"Dataset: {dataset_name}")
        print(f"Sorting by {col}")
        print(final_df)
        print()

        final_df.to_csv(os.path.join(args.output, f"{dataset_name}_{col}.csv"), index=False)



"""
Utility module for CORE

Contains utility functions for:
- Image transformations
- Point set operations
- Deformation field handling
- MHA file operations
"""

from core.utils.util import (
    skip_subsample,
    gamma_corrections,
    tensor_to_rgb_numpy,
    create_deformation_field,
    create_nonrigid_mha,
    create_deform,
    apply_deformation_to_points,
    create_displacement_field_for_wsi,
    _check_dims,
    compute_center_of_mass,
    apply_affine_transformation,
    convert_pytorch_checkpoint,
    create_pyramid,
    scale_transformation_matrix,
    warp_image,
    matrix_mha,
    combine_deformation,
    apply_displacement_field,
    sort_coordinates,
    dice,
    mse,
    rotate_point,
    phase_correlation,
    resample_tensor_to_size,
    gaussian_smoothing_tensor,
    gaussian_smoothing_numpy,
)

from core.utils.mha_wsi import (
    get_image_shape,
    array_to_vips,
    vips_to_array,
    resize_displacement_field,
)

__all__ = [
    # From util.py
    "skip_subsample",
    "gamma_corrections",
    "tensor_to_rgb_numpy",
    "create_deformation_field",
    "create_nonrigid_mha",
    "create_deform",
    "apply_deformation_to_points",
    "create_displacement_field_for_wsi",
    "_check_dims",
    "compute_center_of_mass",
    "apply_affine_transformation",
    "convert_pytorch_checkpoint",
    "create_pyramid",
    "scale_transformation_matrix",
    "warp_image",
    "matrix_mha",
    "combine_deformation",
    "apply_displacement_field",
    "sort_coordinates",
    "dice",
    "mse",
    "rotate_point",
    "phase_correlation",
    "resample_tensor_to_size",
    "gaussian_smoothing_tensor",
    "gaussian_smoothing_numpy",
    
    # From mha_wsi.py
    "get_image_shape",
    "array_to_vips",
    "vips_to_array",
    "resize_displacement_field",
]



import numpy as np
import cv2
import numpy as np
import math
from tiatoolbox.wsicore.wsireader import WSIReader, VirtualWSIReader
import core.preprocessing.tissuemask as tissuemask
import core.preprocessing.preprocessing as prep  # Original preprocessing module

"""
Preprocessing functions for WSI registration
"""
def load_wsi_images(source_path, target_path, resolution=0.625, data='', obj_power=''):
    """
    Load source and target WSI images
    
    Args:
        source_path: Path to source WSI
        target_path: Path to target WSI
        resolution: Resolution for loading
        
    Returns:
        tuple: (source_wsi, target_wsi, source_image, target_image)
    """
    # Load WSI readers
    if data=='anhir':
        target_wsi = VirtualWSIReader.open(target_path)
        source_wsi = VirtualWSIReader.open(source_path)

        # Manually set the objective power if it's None
        if target_wsi.info.objective_power is None:
            target_wsi.info.objective_power = obj_power # Set to a standard value like 20x or 40x
        if source_wsi.info.objective_power is None:
            source_wsi.info.objective_power = obj_power
        
        target = target_wsi.slide_thumbnail(resolution=resolution, units="power")
        source = source_wsi.slide_thumbnail(resolution=resolution, units="power")
        
    else:
        source_wsi = WSIReader.open(source_path)
        target_wsi = WSIReader.open(target_path)
        
        # Load images at specified resolution
        source = load_slide(source_path, resolution)
        target = load_slide(target_path, resolution)
        
        print(f"Source original shape: {source.shape}")
        print(f"Target original shape: {target.shape}")
    
    return source_wsi, target_wsi, source, target


def preprocess_images(source, target):
    """
    Preprocess source and target images
    
    Args:
        source: Source image array
        target: Target image array
        
    Returns:
        tuple: (source_prep, target_prep)
    """
 
    source_prep, target_prep = source, target
    
    print(f"Source preprocessed shape: {source_prep.shape}")
    print(f"Target preprocessed shape: {target_prep.shape}")
    
    return source_prep, target_prep


def extract_tissue_masks(source_prep, target_prep, artefacts):
    """
    Extract tissue masks from preprocessed images
    
    Args:
        source_prep: Preprocessed source image
        target_prep: Preprocessed target image
        
    Returns:
        tuple: (source_mask, target_mask)
    """
    extractor = tissuemask.FlorenceTissueMaskExtractor()
    source_mask = extractor.extract(source_prep, artefacts)
    target_mask = extractor.extract(target_prep, artefacts)
    
    return source_mask, target_mask


def adjust_gamma(image, gamma=0.5):
    """
    Apply gamma correction to an image
    
    Args:
        image: Input image
        gamma: Gamma value
        
    Returns:
        Gamma corrected image
    """
    inv_gamma = 1.0 / gamma
    table = np.array([((i / 255.0) ** inv_gamma) * 255 
                     for i in np.arange(0, 256)]).astype("uint8")
    return cv2.LUT(image, table)


def scale_transformation_matrix(transform_matrix, input_res, output_res):
    """
    Scale transformation matrix to different resolution
    
    Args:
        transform_matrix: Input transformation matrix
        input_res: Input resolution
        output_res: Output resolution
        
    Returns:
        Scaled transformation matrix
    """
    scale_factor = output_res / input_res
    transform_scaled = transform_matrix.copy()
    transform_scaled[0:2, 2] = transform_scaled[0:2, 2] * scale_factor
    return transform_scaled



def load_slide(image_path_1: str, resolution: float = 0.625):
    fixed_wsi = WSIReader.open(image_path_1)
    fixed_thumbnail = fixed_wsi.slide_thumbnail(resolution=resolution, units="power")

    return fixed_thumbnail


def pad_single(image, new_shape):
    y_size, x_size,_ = image.shape
    y_pad = ((int(np.floor((new_shape[0] - y_size)/2))), int(np.ceil((new_shape[0] - y_size)/2)))
    x_pad = ((int(np.floor((new_shape[1] - x_size)/2))), int(np.ceil((new_shape[1] - x_size)/2)))
    new_image = np.pad(image, (y_pad, x_pad,(0,0)), constant_values=0)
    return new_image

def pad_images_np(source, target):
    y_size_source, x_size_source,_ = source.shape
    y_size_target, x_size_target,_ = target.shape
    new_y_size = max(y_size_source, y_size_target)
    new_x_size = max(x_size_source, x_size_target)
    new_shape = (new_y_size, new_x_size)
 
    padded_source = pad_single(source, new_shape)
    padded_target = pad_single(target, new_shape)
    return padded_source, padded_target

def pad_to_same_size(image_1: np.ndarray, image_2: np.ndarray, pad_value: float = 1.0):
    """
    Pad two images to the same size.
    
    Args:
        image_1: First image array with shape (height, width, channel), (batch, height, width, channel), 
                 (height, width), or (batch, channel, height, width)
        image_2: Second image array with shape (height, width, channel), (batch, height, width, channel), 
                 (height, width), or (batch, channel, height, width)
        pad_value: Value to use for padding
        
    Returns:
        tuple: (padded_image_1, padded_image_2, padding_params)
    """
    # Determine the dimensionality and shape of each image
    if image_1.ndim == 4:  # (batch, channel, height, width) format
        y_size_1, x_size_1 = image_1.shape[2], image_1.shape[3]
        y_size_2, x_size_2 = image_2.shape[2], image_2.shape[3]
    elif image_1.ndim == 3:  # (height, width, channel) format - typical RGB image
        y_size_1, x_size_1 = image_1.shape[0], image_1.shape[1]
        y_size_2, x_size_2 = image_2.shape[0], image_2.shape[1]
    else:  # (height, width) format - grayscale image
        y_size_1, x_size_1 = image_1.shape
        y_size_2, x_size_2 = image_2.shape
    
    pad_1 = [(0, 0), (0, 0)]
    pad_2 = [(0, 0), (0, 0)]
    
    # Handle height padding
    if y_size_1 > y_size_2:
        pad_size = y_size_1 - y_size_2
        pad = (math.floor(pad_size / 2), math.ceil(pad_size / 2))
        pad_2[0] = pad
    elif y_size_1 < y_size_2:
        pad_size = y_size_2 - y_size_1
        pad = (math.floor(pad_size / 2), math.ceil(pad_size / 2))
        pad_1[0] = pad
    
    # Handle width padding
    if x_size_1 > x_size_2:
        pad_size = x_size_1 - x_size_2
        pad = (math.floor(pad_size / 2), math.ceil(pad_size / 2))
        pad_2[1] = pad
    elif x_size_1 < x_size_2:
        pad_size = x_size_2 - x_size_1
        pad = (math.floor(pad_size / 2), math.ceil(pad_size / 2))
        pad_1[1] = pad
    
    # Create padded arrays based on dimensionality
    if image_1.ndim == 4:  # (batch, channel, height, width)
        # For 4D tensors: add padding for height and width dimensions
        padded_image_1 = np.pad(
            image_1,
            ((0, 0), (0, 0), pad_1[0], pad_1[1]),
            mode='constant',
            constant_values=pad_value
        )
        padded_image_2 = np.pad(
            image_2,
            ((0, 0), (0, 0), pad_2[0], pad_2[1]),
            mode='constant',
            constant_values=pad_value
        )
    elif image_1.ndim == 3:  # (height, width, channel) - RGB images
        # For RGB images: pad height and width, preserve channels
        padded_image_1 = np.pad(
            image_1,
            (pad_1[0], pad_1[1], (0, 0)),
            mode='constant',
            constant_values=pad_value
        )
        padded_image_2 = np.pad(
            image_2,
            (pad_2[0], pad_2[1], (0, 0)),
            mode='constant',
            constant_values=pad_value
        )
    else:  # (height, width) - grayscale images
        # For 2D arrays
        padded_image_1 = np.pad(
            image_1,
            pad_1,
            mode='constant',
            constant_values=pad_value
        )
        padded_image_2 = np.pad(
            image_2,
            pad_2,
            mode='constant',
            constant_values=pad_value
        )
    
    padding_params = {
        'pad_1': pad_1,
        'pad_2': pad_2
    }
    padded_image_1=gamma_corrections(padded_image_1,1)
    padded_image_2=gamma_corrections(padded_image_2,0.4)
    print("gamma corrected")
    return padded_image_1, padded_image_2, padding_params


def resize_and_compute_translation( moving_image,fixed_image):
    """
    Resizes fixed and moving images to the maximum dimensions using black padding
    and computes initial translation offsets for 2D or 3D images (where the 3rd dimension is the channel).
    Args:
        fixed_image (np.ndarray): The fixed image (2D or 3D).
        moving_image (np.ndarray): The moving image (2D or 3D).
    Returns:
        fixed_padded (np.ndarray): The fixed image with padding.
        moving_padded (np.ndarray): The moving image with padding.
        translation (tuple): The translation offsets (tx, ty) for 2D or (tx, ty, channels) for 3D.
    """
    # Ensure input images are in uint8 format (scaling to 0-255 range if necessary)
    if fixed_image.dtype != np.uint8:
        fixed_image = (fixed_image * 255).astype(np.uint8)
    if moving_image.dtype != np.uint8:
        moving_image = (moving_image * 255).astype(np.uint8)
    
    # Check for 2D or 3D images (where 3D is understood as height, width, channels)
    if fixed_image.ndim == 2:  # 2D images
        fixed_h, fixed_w = fixed_image.shape
        moving_h, moving_w = moving_image.shape
        
        # Compute padding and translation offsets
        max_h, max_w = max(fixed_h, moving_h), max(fixed_w, moving_w)
        if moving_h > fixed_h:
            tx, ty = (max_w - moving_w) // 2, (max_h - moving_h) // 2
        fx, fy = max_w - fixed_w, max_h - fixed_h
        mx, my = max_w - moving_w, max_h - moving_h
        
        # Padding images to match max dimensions
        fixed_padded = np.zeros((max_h, max_w), dtype=np.uint8)
        moving_padded = np.zeros((max_h, max_w), dtype=np.uint8)
        fixed_padded[:fixed_h, :fixed_w] = fixed_image
        # moving_padded[ty:ty + moving_h, tx:tx + moving_w] = moving_image
        moving_padded[:moving_h, :moving_w] = moving_image
        
        return fixed_padded, moving_padded, (fx, fy), (mx, my)
    
    elif fixed_image.ndim == 3:  # 3D images (height, width, channels)
        fixed_h, fixed_w, fixed_c = fixed_image.shape
        moving_h, moving_w, moving_c = moving_image.shape
        
        # Compute padding and translation offsets for height, width (not channels)
        max_h, max_w = max(fixed_h, moving_h), max(fixed_w, moving_w)
        tx, ty = (max_w - moving_w) // 2, (max_h - moving_h) // 2
        fx, fy = max_w - fixed_w, max_h - fixed_h
        mx, my = max_w - moving_w, max_h - moving_h
        
        # Padding images to match max height and width (but retain the channel dimension)
        fixed_padded = np.zeros((max_h, max_w, fixed_c), dtype=np.uint8)
        moving_padded = np.zeros((max_h, max_w, moving_c), dtype=np.uint8)
        moving_padded[:moving_h, :moving_w] = moving_image
        fixed_padded[:fixed_h, :fixed_w, :] = fixed_image
        # moving_padded[ty:ty + moving_h, tx:tx + moving_w, :] = moving_image
        fixed_padded=gamma_corrections( fixed_padded,1)
        moving_padded=gamma_corrections(moving_padded,0.4)
        return fixed_padded, moving_padded, (fx, fy), (mx, my)
    
    else:
        raise ValueError("Input images must be either 2D or 3D with channel as the 3rd dimension.") 
def gamma_corrections(img, gamma):
        invGamma = 1.0 / gamma
        table = np.array([((i / 255.0) ** invGamma) * 255 for i in np.arange(0, 256)]).astype("uint8")
        return cv2.LUT(img, table)
def process_nuclei_patch(img, threshold, gamma=None, min_area=200):
    """
    Process a single patch to detect nuclei
    
    Args:
        img: Input image patch
        threshold: Binary threshold value
        gamma: Gamma correction value (optional)
        min_area: Minimum area for nuclei detection
        
    Returns:
        tuple: (binary_image, stats, centroids)
    """
    # Convert to grayscale
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    
    # Apply gamma correction if specified
    if gamma is not None:
        gray = adjust_gamma(gray, gamma)
    
    # Apply binary threshold
    _, binary = cv2.threshold(gray, threshold, 255, cv2.THRESH_BINARY_INV)
    
    # Find connected components
    num_labels, labels, stats, centroids = cv2.connectedComponentsWithStats(binary)
    
    return binary, stats, centroids



"""
Evaluation module for CORE

Contains evaluation metrics and functions for:
- Target Registration Error (TRE)
- Normalized Gradient Field (NGF) metric
- Landmark-based evaluation
- Displacement field evaluation
"""

from core.evaluation.evaluation import (
    tre,
    rtre,
    load_landmark_points,
    load_evaluation_landmarks,
    transform_points_homogeneous,
    evaluate_registration_tre,
    apply_displacement_field_to_points,
    evaluate_nonrigid_registration,
    ngf_metric,
)

__all__ = [
    "tre",
    "rtre",
    "load_landmark_points",
    "load_evaluation_landmarks",
    "transform_points_homogeneous",
    "evaluate_registration_tre",
    "apply_displacement_field_to_points",
    "evaluate_nonrigid_registration",
    "ngf_metric",
]



"""
    "XFeat: Accelerated Features for Lightweight Image Matching, CVPR 2024."
    https://www.verlab.dcc.ufmg.br/descriptors/xfeat_cvpr24/

    This script implements color + geometric transformations using Kornia.
    Given a dataset of random real unlabeled images, we apply photometric transformations, 
    homography warps and also TPS warps. It also handles black borders by
    pasting a random background image.
"""

import torch
from torch import nn
from torch.utils.data import Dataset
import torch.utils.data as data
from torchvision import transforms
import torch.nn.functional as F

import cv2
import kornia
import kornia.augmentation as K
from kornia.geometry.transform import get_tps_transform as findTPS
from kornia.geometry.transform import warp_points_tps, warp_image_tps

import glob
import random
import tqdm

import numpy as np
import pdb
import time

random.seed(0)
torch.manual_seed(0)

def generateRandomTPS(shape, grid = (8, 6), GLOBAL_MULTIPLIER = 0.3, prob = 0.5):

    h, w = shape
    sh, sw = h/grid[0], w/grid[1]
    src = torch.dstack(torch.meshgrid(torch.arange(0, h + sh , sh),
                         torch.arange(0, w + sw , sw), indexing='ij'))

    offsets = torch.rand(grid[0]+1, grid[1]+1, 2) - 0.5
    offsets *= torch.tensor([ sh/2, sw/2 ]).view(1, 1, 2)  * min(0.97, 2.0 * GLOBAL_MULTIPLIER)
    dst = src + offsets if np.random.uniform() < prob else src
    
    src, dst = src.view(1, -1, 2), dst.view(1, -1, 2)
    src = (src / torch.tensor([h,w]).view(1,1,2) ) * 2 - 1.
    dst = (dst / torch.tensor([h,w]).view(1,1,2) ) * 2 - 1.
    weights, A = findTPS(dst, src)

    return src, weights, A


def generateRandomHomography(shape, GLOBAL_MULTIPLIER = 0.3):
    #Generate random in-plane rotation [-theta,+theta]
    theta = np.radians(np.random.uniform(-30, 30))

    #Generate random scale in both x and y
    scale_x, scale_y = np.random.uniform(0.35, 1.2, 2)

    #Generate random translation shift
    tx , ty = -shape[1]/2.0 , -shape[0]/2.0 
    txn, tyn = np.random.normal(0, 120.0*GLOBAL_MULTIPLIER, 2) 

    c, s = np.cos(theta), np.sin(theta)

    #Affine coeffs
    sx , sy = np.random.normal(0,0.6*GLOBAL_MULTIPLIER,2)

    #Projective coeffs
    p1 , p2 = np.random.normal(0,0.006*GLOBAL_MULTIPLIER,2)


    # Build Homography from parmeterizations
    H_t = np.array(((1,0, tx), (0, 1, ty), (0,0,1))) #t

    H_r = np.array(((c,-s, 0), (s, c, 0), (0,0,1))) #rotation,
    H_a = np.array(((1,sy, 0), (sx, 1, 0), (0,0,1))) # affine
    H_p = np.array(((1, 0, 0), (0 , 1, 0), (p1,p2,1))) # projective

    H_s = np.array(((scale_x,0, 0), (0, scale_y, 0), (0,0,1))) #scale
    H_b = np.array(((1.0,0,-tx +txn), (0, 1, -ty + tyn), (0,0,1))) #t_back,

    #H = H_e * H_s * H_a * H_p
    H = np.dot(np.dot(np.dot(np.dot(np.dot(H_b,H_s),H_p),H_a),H_r),H_t)

    return H


class AugmentationPipe(nn.Module):
    def __init__(
                self, device, load_dataset = True,
                img_dir = "/homeLocal/guipotje/sshfs/datasets/coco_20k",
                warp_resolution = (1200, 900),
                out_resolution = (400, 300),
                sides_crop = 0.2,
                max_num_imgs = 50,
                num_test_imgs = 10,
                batch_size = 1,
                photometric = True,
                geometric = True,
                reload_step = 1_000
                ):
        super(AugmentationPipe, self).__init__()
        self.half = 16
        self.device = device

        self.dims = warp_resolution
        self.batch_size = batch_size
        self.out_resolution = out_resolution 
        self.sides_crop = sides_crop
        self.max_num_imgs = max_num_imgs
        self.num_test_imgs = num_test_imgs
        self.dims_t = torch.tensor([int(self.dims[0]*(1. - self.sides_crop)) - int(self.dims[0]*self.sides_crop) -1,
                                    int(self.dims[1]*(1. - self.sides_crop)) - int(self.dims[1]*self.sides_crop) -1]).float().to(device).view(1,1,2)
        self.dims_s = torch.tensor([ self.dims_t[0,0,0] / out_resolution[0],
                                     self.dims_t[0,0,1] / out_resolution[1]]).float().to(device).view(1,1,2) 

        self.all_imgs = glob.glob(img_dir + '/*.jpg') + glob.glob(img_dir + '/*.png')
        
        self.photometric = photometric
        self.geometric = geometric
        self.cnt = 1
        self.reload_step = reload_step

        list_augmentation = [
                        #kornia.augmentation.RandomChannelShuffle(p=0.5),
                        kornia.augmentation.ColorJitter(0.15, 0.15, 0.15, 0.15, p=1.),
                        kornia.augmentation.RandomEqualize(p = 0.4),
                        kornia.augmentation.RandomGaussianBlur(p = 0.3, sigma = (2.0, 2.0), kernel_size = (7,7))
                        ]
        if photometric is False:
            list_augmentation = []

        self.aug_list = kornia.augmentation.ImageSequential(*list_augmentation)
        
        if len(self.all_imgs) < 10:
            raise RuntimeError('Couldnt find enough images to train. Please check the path: ', img_dir)

        if load_dataset:
            print('[Synthetic] Found a total of ', len(self.all_imgs), ' images for training..')

            if len(self.all_imgs) - num_test_imgs < max_num_imgs:
                raise RuntimeError('Error: test set overlaps with training set! Decrease number of test imgs')

            self.load_imgs()

            self.TPS = True


    def load_imgs(self):

            random.shuffle(self.all_imgs)

            train = []
            fast = cv2.FastFeatureDetector_create(30)
            for p in tqdm.tqdm(self.all_imgs[:self.max_num_imgs], desc='loading train'):
                im = cv2.imread(p)
                halfH, halfW = im.shape[0]//2, im.shape[1]//2
                if halfH > halfW:
                    im = np.rot90(im)
                    halfH, halfW = halfW, halfH

                if im.shape[0] != self.dims[1] or im.shape[1] != self.dims[0]:
                    im = cv2.resize(im, self.dims)

                train.append(np.copy(im))

            self.train = train
            
            self.test = [
                        cv2.resize(cv2.imread(p), self.dims)                 
                        for p in tqdm.tqdm(self.all_imgs[-self.num_test_imgs:],
                                           desc='loading test')
                        ] 

    def norm_pts_grid(self, x):
        if len(x.size()) == 2:
            return (x.view(1,-1,2) * self.dims_s / self.dims_t) * 2. - 1 
        return (x * self.dims_s / self.dims_t) * 2. - 1

    def denorm_pts_grid(self, x):
        if len(x.size()) == 2:
            return ((x.view(1,-1,2) + 1) / 2.) / self.dims_s * self.dims_t
        return ((x+1) / 2.) / self.dims_s * self.dims_t

    def rnd_kps(self, shape, n = 256):
        h, w = shape
        kps = torch.rand(size = (3,n)).to(self.device)
        kps[0,:]*=w
        kps[1,:]*=h
        kps[2,:] = 1.0

        return kps

    def warp_points(self, H, pts):
      scale = self.dims_s.view(-1,2)
      offset = torch.tensor([int(self.dims[0]*self.sides_crop), int(self.dims[1]*self.sides_crop)], device = pts.device).float()
      pts = pts*scale + offset
      pts = torch.vstack( [pts.t(), torch.ones(1, pts.shape[0], device = pts.device)])
      warped = torch.matmul(H, pts)
      warped = warped / warped[2,...] 
      warped = warped.t()[:, :2]
      return (warped - offset) / scale

    @torch.inference_mode()
    def forward(self, x, difficulty = 0.3, TPS = False, prob_deformation = 0.5, test = False):
        """
            Perform augmentation to a batch of images.

            input:
                x -> torch.Tensor(B, C, H, W): rgb images
                difficulty -> float: level of difficulty, 0.1 is medium, 0.3 is already pretty hard
                tps -> bool: Wether to apply non-rigid deformations in images
                prob_deformation -> float: probability to apply a deformation

            return:
                'output'    ->   torch.Tensor(B, C, H, W): rgb images
                Tuple:
                    'H'       ->   torch.Tensor(3,3): homography matrix 
                    'mask'  ->     torch.Tensor(B, H, W): mask of valid pixels after warp
                    (deformation only)
                    src, weights, A are parameters from a TPS warp (all torch.Tensors)

        """

        if self.cnt % self.reload_step == 0:
            self.load_imgs()

        if self.geometric is False:
            difficulty = 0.

        with torch.no_grad():
            x = (x/255.).to(self.device)
            b, c, h, w = x.shape
            shape = (h, w)
              
            ######## Geometric Transformations

            H = torch.tensor(np.array([generateRandomHomography(shape, difficulty) for b in range(self.batch_size)]),
                               dtype = torch.float32).to(self.device)
            
            output = kornia.geometry.transform.warp_perspective(x, H,
                            dsize = shape, padding_mode = 'zeros')

            #crop % of image boundaries each side to reduce invalid pixels after warps
            low_h = int(h * self.sides_crop); low_w = int(w*self.sides_crop)
            high_h = int(h*(1. - self.sides_crop)); high_w= int(w * (1. - self.sides_crop))
            output = output[..., low_h:high_h, low_w:high_w]
            x = x[..., low_h:high_h, low_w:high_w]

            #apply TPS if desired:
            if TPS:
                src, weights, A = None, None, None
                for b in range(self.batch_size):
                    b_src, b_weights, b_A = generateRandomTPS(shape, (8,6), difficulty, prob = prob_deformation)
                    b_src, b_weights, b_A = b_src.to(self.device), b_weights.to(self.device), b_A.to(self.device)

                    if src is None:
                        src, weights, A = b_src, b_weights, b_A
                    else:
                        src = torch.cat((b_src, src))
                        weights = torch.cat((b_weights, weights))
                        A = torch.cat((b_A, A))

                output = warp_image_tps(output, src, weights, A)

            output = F.interpolate(output, self.out_resolution[::-1], mode = 'nearest')
            x = F.interpolate(x, self.out_resolution[::-1], mode = 'nearest')

            mask = ~torch.all(output == 0, dim=1, keepdim=True)
            mask = mask.expand(-1,3,-1,-1)

            # Make-up invalid regions with texture from the batch
            rv = 1 if not TPS else 2
            output_shifted = torch.roll(x, rv, 0)
            output[~mask] = output_shifted[~mask]
            mask = mask[:, 0, :, :]

            ######## Photometric Transformations
            output = self.aug_list(output)

            b, c, h, w = output.shape
            #Correlated Gaussian Noise
            if np.random.uniform() > 0.5 and self.photometric:
                noise = F.interpolate(torch.randn_like(output)*(10/255), (h//2, w//2))
                noise = F.interpolate(noise, (h, w), mode = 'bicubic')
                output = torch.clip( output + noise, 0., 1.)

            #Random shadows
            if np.random.uniform() > 0.6 and self.photometric:
                noise = torch.rand((b, 1, h//64, w//64), device = self.device) * 1.3
                noise = torch.clip(noise, 0.25, 1.0)
                noise = F.interpolate(noise, (h, w), mode = 'bicubic')
                noise = noise.expand(-1, 3, -1, -1)
                output *= noise
                output = torch.clip( output, 0., 1.)

            self.cnt+=1

        if TPS:
            return output, (H, src, weights, A, mask)
        else:
            return output, (H, mask)

    def get_correspondences(self, kps_target, T):
        H, H2, src, W, A = T
        undeformed  = self.denorm_pts_grid(   
                                        warp_points_tps(self.norm_pts_grid(kps_target),
                                        src, W, A) ).view(-1,2)

        warped_to_src = self.warp_points(H@torch.inverse(H2), undeformed)

        return warped_to_src


#  Accelerated feature model from XFeat

dependencies = ['torch']
from modules.xfeat import XFeat as _XFeat
import torch

def XFeat(pretrained=True, top_k=4096, detection_threshold=0.05):
    """
    XFeat model
    pretrained (bool): kwargs, load pretrained weights into the model
    """
    weights = None
    if pretrained:
        weights = torch.hub.load_state_dict_from_url("https://github.com/verlab/accelerated_features/raw/main/weights/xfeat.pt")
    
    model = _XFeat(weights, top_k=top_k, detection_threshold=detection_threshold)
    return model



import os
import cv2
import math
import pandas as pd
import matplotlib.pyplot as plt
import SimpleITK as sitk
import csv
import torch
import torch.nn.functional as F
import torchvision.transforms as transforms
import math
import numpy as np
from skimage import exposure, filters, img_as_float, color, measure, morphology
from skimage.registration import phase_cross_correlation
from sklearn.neighbors import NearestNeighbors
from sklearn.cluster import KMeans
from scipy import ndimage as nd
from scipy import ndimage as scnd
from scipy.ndimage import map_coordinates, gaussian_filter
from tiatoolbox.utils.metrics import dice
from tiatoolbox import logger, rcParam
from scipy.interpolate import griddata

RGB_IMAGE_DIM = 3
BIN_MASK_DIM = 2


# ---------------------- Tensor Utility Functions ---------------------- #

def gaussian_smoothing_tensor(input_tensor: torch.Tensor, blur_sigma: float) -> torch.Tensor:
    """Apply Gaussian smoothing to a tensor."""
    with torch.set_grad_enabled(False):
        kernel_width = int(blur_sigma * 2.54) + 1 
        if kernel_width % 2 == 0:
            kernel_width += 1
        return transforms.GaussianBlur(kernel_width, blur_sigma)(input_tensor)


def resample_tensor_to_size(input_tensor: torch.Tensor, 
                            target_dimensions: torch.Size, 
                            mode: str = 'bilinear') -> torch.Tensor:
    """Resize tensor to target dimensions using interpolation."""
    return F.interpolate(input_tensor, size=target_dimensions, 
                        mode=mode, align_corners=False)


def gaussian_smoothing_numpy(image: np.ndarray, sigma: float) -> np.ndarray:
    """Apply Gaussian smoothing to a numpy array."""
    return gaussian_filter(image.astype(np.float32), sigma=sigma)


#  UTILS
def skip_subsample(points, n_samples=1000):
    total_points = points.shape[0]
    if total_points <= n_samples:
        return points
    step = total_points // n_samples
    return points[::step][:n_samples]

def gamma_corrections(img, gamma):
        invGamma = 1.0 / gamma
        table = np.array([((i / 255.0) ** invGamma) * 255 for i in np.arange(0, 256)]).astype("uint8")
        return cv2.LUT(img, table)
         
def tensor_to_rgb_numpy(tensor):
    # (1, 1, H, W) -> (3, H, W) -> (H, W, 3)
    tensor_rgb = tensor.squeeze().repeat(3, 1, 1)
    return tensor_rgb.permute(1, 2, 0).detach().cpu().numpy()
    
# DEFORMATION
def create_deformation_field(shape_transform, source_prep, u_x, u_y, util, output_path='./533_finerigid.mha'):
    """
    Creates a deformation field by scaling a 3x3 transformation matrix and 
    composing it with given vector fields.

    Parameters
    ----------
    shape_transform : np.ndarray
        A 3x3 transformation matrix (affine or rigid).
    source_prep : np.ndarray
        Source coordinate grid or image array used for transformation.
    u_x, u_y : np.ndarray
        Components of an existing deformation field (x and y displacement maps).
    util : module
        A module or object providing two functions:
            - matrix_df(source, transform): applies rigid transformation.
            - combine_deformation(u_x, u_y, t_x, t_y): composes deformation fields.
    output_path : str, optional
        Path to save the resulting deformation field as an .mha file.

    Returns
    -------
    sitk.Image
        The resulting deformation field as a SimpleITK image.
    """

    # Define scale factor
    scale_factor = 0.625 / 40  # equivalent to 0.015625

    # Scale the translation components of the transformation matrix
    transform_3x3_scaled = shape_transform.copy()
    transform_3x3_scaled[0:2, 2] *= scale_factor

    # Apply inverse of scaled transform
    sh_transform = transform_3x3_scaled
    t_x, t_y = matrix_df(source_prep, np.linalg.inv(sh_transform))

    # Compose the new vector field
    f_x, f_y = combine_deformation(u_x, u_y, t_x, t_y)

    # Stack into deformation field
    deformation_field = np.stack((f_x, f_y), axis=-1)

    # Convert to SimpleITK image and save
    sitk_image = sitk.GetImageFromArray(deformation_field)
    sitk.WriteImage(sitk_image, output_path)

    return sitk_image
def create_nonrigid_mha(
    moving_subsample,
    nonrigid_transformed_coords,
    r_x,
    w_x, w_y,
    target_prep,
    create_displacement_field,
    RegistrationParams,
    output_path=""
):
    """
    Builds a non-rigid displacement + deformation field 
    and saves it as an .mha file (using fr_x, fr_y as final field).
    """

    print("Creating displacement field...")

    # Scale for numerical stability
    scale_factor = 64
    source_points_scaled = moving_subsample / scale_factor
    target_points_scaled = nonrigid_transformed_coords / scale_factor

    # Determine grid size
    H, W = r_x.shape
    grid_y, grid_x = np.mgrid[0:H, 0:W]

    # Dense displacement field
    displacement_field = create_displacement_field(
        source_points_scaled,
        target_points_scaled,
        target_prep.shape,
        method=RegistrationParams.INTERPOLATION_METHOD,
        sigma=RegistrationParams.DISPLACEMENT_SIGMA,
        max_displacement=RegistrationParams.MAX_DISPLACEMENT
    )

    # Combine deformation + displacement
    fr_x, fr_y = util.combine_deformation(
        w_x, w_y,
        displacement_field[..., 0],
        displacement_field[..., 1]
    )

    # FINAL deformation field to save (now fr_x, fr_y)
    deformation_field = np.stack((fr_x, fr_y), axis=-1)

    # Convert → SimpleITK image
    sitk_image = sitk.GetImageFromArray(deformation_field)

    # Save to disk
    sitk.WriteImage(sitk_image, output_path)

    return sitk_image, deformation_field, displacement_field, (fr_x, fr_y)

def create_deform(source_prep, final_transform, displacement_field, output_path=""):
    """
    Computes the full deformation field and writes it to an MHA file.

    Parameters
    ----------
    source_prep : array-like
        Input source data for matrix_df.
    final_transform : array-like
        Final transform data for matrix_df.
    displacement_field : array-like
        Raw displacement field (expected shape: [2, H, W] or similar).
    output_path : str
        Output MHA file path.
    """
    
    # Get rigid or base transform components
    r_x, r_y = matrix_df(source_prep, final_transform)

    # Convert displacement field into 2-component arrays
    disp_field = deform_conversion(displacement_field)

    # Combine base transform + deformation
    w_x, w_y = combine_deformation(r_x, r_y, disp_field[0], disp_field[1])

    # Stack into deformation field (H, W, 2)
    deformation_field = np.stack((w_x, w_y), axis=-1)

    # Convert to SimpleITK image
    sitk_image = sitk.GetImageFromArray(deformation_field)

    # Write to file
    sitk.WriteImage(sitk_image, output_path)

    return sitk_image
def apply_deformation_to_points(points, deformation_field):
    """
    Apply a deformation field to a set of 2D points.

    Args:
        points: np.array of shape (N, 2), points as (x, y) coordinates.
        deformation_field: np.array of shape (2, H, W), displacement vectors.
                           deformation_field[0] is displacement in x,
                           deformation_field[1] is displacement in y.

    Returns:
        warped_points: np.array of shape (N, 2), warped point coordinates.
    """
    # Extract deformation components
    disp_x = deformation_field[0]
    disp_y = deformation_field[1]

    H, W = disp_x.shape
    print("height", H)
    print("width", W)
    # Points might be float and anywhere inside image, so use interpolation of deformation field
    # Interpolate displacement at each point's coordinate
    warped_points = np.zeros_like(points)

    # Points are (x, y), but map_coordinates requires (row, col) = (y, x)
    # So we need to query displacement fields at (y, x)
    coords = np.array([points[:,1], points[:,0]])  # shape (2, N)

    # Interpolate displacement fields at these coords
    disp_at_points_x = map_coordinates(disp_x, coords, order=1, mode='nearest')
    disp_at_points_y = map_coordinates(disp_y, coords, order=1, mode='nearest')

    # Apply displacement: new_pos = original_pos + displacement
    warped_points[:, 0] = points[:, 0] + disp_at_points_x
    warped_points[:, 1] = points[:, 1] + disp_at_points_y

    return warped_points
def create_displacement_field_for_wsi(transform_matrix, source_thumbnail, target_thumbnail):    
    # Use the larger dimensions to avoid cropping issues
    max_height = max(source_thumbnail.shape[0], target_thumbnail.shape[0])
    max_width = max(source_thumbnail.shape[1], target_thumbnail.shape[1])
    
    # Create coordinate grid at the chosen dimensions
    y_coords, x_coords = np.mgrid[0:max_height, 0:max_width]
    
    # Convert to homogeneous coordinates for transformation
    coords_homogeneous = np.stack([
        x_coords.flatten(), 
        y_coords.flatten(), 
        np.ones(x_coords.size)
    ], axis=0)
    
    transform_inv = np.linalg.inv(transform_matrix)
    transformed_coords = transform_inv @ coords_homogeneous
    
    # Reshape back to 2D grids
    source_x = transformed_coords[0].reshape(max_height, max_width)
    source_y = transformed_coords[1].reshape(max_height, max_width)
    
    u_x = source_x - x_coords
    u_y = source_y - y_coords
    # Stack into displacement field (H, W, 2)
    displacement_field = np.stack((u_x, u_y), axis=-1)
    return displacement_field
# POINT


def _check_dims(
    fixed_img: np.ndarray,
    moving_img: np.ndarray,
    fixed_mask: np.ndarray,
    moving_mask: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    """Check the dimensionality of images and mask.

    This function verify the dimensionality of images and their corresponding masks.
    If the input images are RGB images, it converts them to grayscale images.

    Args:
        fixed_img (:class:`numpy.ndarray`):
            A fixed image.
        moving_img (:class:`numpy.ndarray`):
            A moving image.
        fixed_mask (:class:`numpy.ndarray`):
            A binary tissue mask for the fixed image.
        moving_mask (:class:`numpy.ndarray`):
            A binary tissue mask for the moving image.

    Returns:
        tuple:
            - :class:`numpy.ndarray` - A grayscale fixed image.
            - :class:`numpy.ndarray` - A grayscale moving image.

    """
    if len(np.unique(fixed_mask)) == 1 or len(np.unique(moving_mask)) == 1:
        msg = "The foreground is missing in the mask."
        raise ValueError(msg)

    if (
        fixed_img.shape[:2] != fixed_mask.shape
        or moving_img.shape[:2] != moving_mask.shape
    ):
        msg = "Mismatch of shape between image and its corresponding mask."
        raise ValueError(msg)

    if len(fixed_img.shape) == RGB_IMAGE_DIM:
        fixed_img = cv2.cvtColor(fixed_img, cv2.COLOR_BGR2GRAY)

    if len(moving_img.shape) == RGB_IMAGE_DIM:
        moving_img = cv2.cvtColor(moving_img, cv2.COLOR_BGR2GRAY)

    return fixed_img, moving_img



def compute_center_of_mass(mask: np.ndarray) -> tuple:
    """Compute center of mass.

    Args:
        mask: (:class:`numpy.ndarray`):
            A binary mask.

    Returns:
        :py:obj:`tuple` - x- and y- coordinates representing center of mass.
            - :py:obj:`int` - X coordinate.
            - :py:obj:`int` - Y coordinate.

    """
    moments = cv2.moments(mask)
    x_coord_center = moments["m10"] / moments["m00"]
    y_coord_center = moments["m01"] / moments["m00"]
    return (x_coord_center, y_coord_center)


def apply_affine_transformation(
    fixed_img: np.ndarray,
    moving_img: np.ndarray,
    transform_initializer: np.ndarray,
) -> np.ndarray:
    """Apply affine transformation using OpenCV.

    Args:
        fixed_img (:class:`numpy.ndarray`):
            A fixed image.
        moving_img (:class:`numpy.ndarray`):
            A moving image.
        transform_initializer (:class:`numpy.ndarray`):
            A rigid transformation matrix.

    Returns:
        :class:`numpy.ndarray`:
            A transformed image.

    Examples:
        >>> moving_image = apply_affine_transformation(
        ...     fixed_image, moving_image, transform_initializer
        ... )

    """
    return cv2.warpAffine(
        moving_img,
        transform_initializer[0:-1][:],
        fixed_img.shape[:2][::-1],
    )
# checkpoint conversion
def convert_pytorch_checkpoint(net_state_dict):
    variable_name_list = list(net_state_dict.keys())
    is_in_parallel_mode = all(v.split(".")[0] == "module" for v in variable_name_list)
    if is_in_parallel_mode:
        net_state_dict = {
            ".".join(k.split(".")[1:]): v for k, v in net_state_dict.items()
        }
    return net_state_dict

def create_pyramid(array: np.ndarray, num_levels: int, mode: str = 'bilinear'):
    """
    Creates a resolution pyramid for a NumPy array (assumes uniform downsampling by a factor of 2).

    Parameters
    ----------
    array : np.ndarray
        The input array (image or multi-dimensional data).
    num_levels : int
        The number of pyramid levels.
    mode : str
        The interpolation mode ("bilinear" or "nearest").

    Returns
    ----------
    pyramid : list of np.ndarray
        The created resolution pyramid.
    """
    pyramid = [None] * num_levels
    for i in range(num_levels - 1, -1, -1):
        if i == num_levels - 1:
            pyramid[i] = array
        else:
            current_size = pyramid[i + 1].shape
            new_size = tuple(int(current_size[j] / 2) if j > 1 else current_size[j] for j in range(len(current_size)))
            new_size = new_size[2:]  # Exclude batch and channel dimensions
            smoothed = gaussian_smoothing_tensor(pyramid[i + 1], 1)
            new_array = resample_tensor_to_size(smoothed, new_size, mode=mode)
            pyramid[i] = new_array
    return pyramid

def scale_transformation_matrix(transform_matrix, input_res, output_res):
    """
    Upscale the transformation matrix to the original image resolution.

    Args:
        transform_matrix (np.ndarray): Transformation matrix.
        resolution (int): Original image resolution.

    Returns:
        np.ndarray: Upscaled transformation matrix.
    """
    scale_factor = output_res / input_res
    transform_upscaled = transform_matrix.copy()
    transform_upscaled[0:2][:, 2] = transform_upscaled[0:2][:, 2] * scale_factor

    return transform_upscaled

def warp_image(image, u_x, u_y):
    y_size, x_size = image.shape
    grid_x, grid_y = np.meshgrid(np.arange(x_size), np.arange(y_size))
    return nd.map_coordinates(image, [grid_y + u_y, grid_x + u_x], order=3, cval=0.0)


def matrix_mha(image, matrix):
    y_size, x_size = np.shape(image)
    x_grid, y_grid = np.meshgrid(np.arange(x_size), np.arange(y_size))
    points = np.vstack((x_grid.ravel(), y_grid.ravel(), np.ones(np.shape(image)).ravel()))
    transformed_points = matrix @ points
    u_x = np.reshape(transformed_points[0, :], (y_size, x_size)) - x_grid
    u_y = np.reshape(transformed_points[1, :], (y_size, x_size)) - y_grid
    return u_x, u_y

def combine_deformation(u_x, u_y, v_x, v_y):
    y_size, x_size = np.shape(u_x)
    print("u_x shape: ", u_x.shape)
    grid_x, grid_y = np.meshgrid(np.arange(x_size), np.arange(y_size))
    added_y = grid_y + v_y
    added_x = grid_x + v_x
    t_x = nd.map_coordinates(grid_x + u_x, [added_y, added_x], mode='constant', cval=0.0)
    t_y = nd.map_coordinates(grid_y + u_y, [added_y, added_x], mode='constant', cval=0.0)
    n_x, n_y = t_x - grid_x, t_y - grid_y
    indexes_x = np.logical_or(added_x >= x_size - 1, added_x <= 0)
    indexes_y = np.logical_or(added_y >= y_size - 1, added_y <= 0)
    indexes = np.logical_or(indexes_x, indexes_y)
    n_x[indexes] = 0.0
    n_y[indexes] = 0.0
    return n_x, n_y

def apply_displacement_field(image, displacement_field):
    """
    Apply displacement field to warp an image.
    
    Args:
        image: Input image
        displacement_field: Displacement field [2, H, W]
        
    Returns:
        warped_image: Warped image
    """
    h, w = image.shape[:2]
    grid_y, grid_x = np.mgrid[0:h, 0:w]
    
    # Add displacement to grid coordinates
    coords_y = grid_y + displacement_field[1]
    coords_x = grid_x + displacement_field[0]
    
    # Stack coordinates for mapping
    coords = np.stack([coords_y, coords_x], axis=0)
    
    # Apply mapping to each channel
    if len(image.shape) == 3:
        warped_image = np.zeros_like(image)
        for c in range(image.shape[2]):
            warped_image[:, :, c] = scnd.map_coordinates(image[:, :, c], coords, order=1)
    else:
        warped_image = scnd.map_coordinates(image, coords, order=1)
    
    return warped_image

# ------------------------
def sort_coordinates(set1, set2):
    """
    Match points between fixed and moving point set.

    Args:
        set1 (np.array): Fixed point set
        set2 (np.array): Moving point set

    Returns:
        np.array: New moving point set
    """
    # Match points using K nearest neighbors.
    sorted_move = []
    set1 = np.array([[coord[0], coord[1]] for coord in set1])
    set2 = np.array([[coord[0], coord[1]] for coord in set2])
    knn = NearestNeighbors(n_neighbors=1, algorithm='auto').fit(set2)
    for coord in set1:
        distances, indices = knn.kneighbors([[coord[0], coord[1]]])
        sorted_move.append(set2[indices])
    # Return matched points.
    return np.array(sorted_move)

def dice(fixed, moving):
    # Calculate DICE coefficient between fixed and moving images.
    intersection = np.sum(np.logical_and(fixed, moving))
    union = np.sum(np.logical_or(fixed, moving))
    
    if union == 0:
        return 0  # handle the case where both arrays are empty
    
    dice_coefficient = 2.0 * intersection / union
    return dice_coefficient

def mse(fixed, moving):
    """
    Calculate MSE value.

    Args:
        fixed (np.array): Fixed point set
        moving (np.array): Moving point set

    Returns:
        float: Mean Square error.
    """
    if len(fixed) == 0 or len(moving) == 0:
        print("Warning: One or both point sets are empty. Returning inf for MSE.")
        return float('inf') 
    else:
        # Match fixed and moving point set.
        moving = sort_coordinates(fixed, moving)
        s = 0
        # Calculate euclidean distances between matched points.
        for f, m in zip(fixed, moving):
            s += (f[0]-m[0][0][0])**2 + (f[1]-m[0][0][1])**2
        # Calculates MSE value and return value
        mse_value = s/len(fixed)

    return mse_value



def rotate_point(x, y, cx, cy, angle):
    """
    Rotate point set around center with certain angle.

    Args:
        x (float): x-coordinate
        y (float): y-coordinate
        cx (float): center x-coordinate
        cy (float): center y-coordinate
        angle (float): rotation angle

    Returns:
        float: rotated x coordinate
        float: rotated y coordinate
    """
    # convert angles to radians
    angle_radians = math.radians(angle)
    
    # Apply rotation formula
    x_prime = (x - cx) * math.cos(angle_radians) - (y - cy) * math.sin(angle_radians) + cx
    y_prime = (x - cx) * math.sin(angle_radians) + (y - cy) * math.cos(angle_radians) + cy
    
    return x_prime, y_prime


    
def phase_correlation(fixed, moving):
    # Run phase correlation shift between point sets as second measure for translation.
    shift, _, _ = phase_cross_correlation(fixed, moving)
    return np.array(shift)


def find_scale(images):
    # Find x and y scale factors between fixed and moving point set using the difference between min and max x and y point sets.
    min_x_move, max_x_move, min_y_move, max_y_move = np.min(images.moving.points[:, 0]), np.max(images.moving.points[:, 0]), np.min(images.moving.points[:, 1]), np.max(images.moving.points[:, 1])
    min_x_fixed, max_x_fixed, min_y_fixed, max_y_fixed = np.min(images.fixed.points[:, 0]), np.max(images.fixed.points[:, 0]), np.min(images.fixed.points[:, 1]), np.max(images.fixed.points[:, 1])
    scale_factor_x = (max_x_fixed-min_x_fixed) / (max_x_move-min_x_move)
    scale_factor_y = (max_y_fixed-min_y_fixed) / (max_y_move-min_y_move)
    return scale_factor_x, scale_factor_y

def scale_coordinates(images):
    # Find scale factors and apply onto moving point sets.
    scale_factor_x , scale_factor_y = find_scale(images)
    scale_factor = max(scale_factor_x, scale_factor_y)
    min_x_move, max_x_move, min_y_move, max_y_move = np.min(images.moving.points[:, 0]), np.max(images.moving.points[:, 0]), np.min(images.moving.points[:, 1]), np.max(images.moving.points[:, 1])
    x_center = (max_x_move+min_x_move) / 2
    y_center = (max_y_move+min_y_move) / 2
    translated_coordinates = [(x - x_center, y - y_center) for x, y in images.moving.points]

    scaled_coordinates = [(scale_factor_x * x, scale_factor_y * y) for x, y in translated_coordinates]

    final_coordinates = np.array([[x + x_center, y + y_center] for x, y in scaled_coordinates])
    # Apply scalings and set point sets.
    images.moving.set_points(final_coordinates)
    return images, 1/scale_factor_x, 1/scale_factor_y


def matchpoints(points):
    """
    Match point between fixed and moving point sets.

    Args:
        points (np.array): point sets

    Returns:
        np.array: Rotation matrix and translation vector
    """
    if not points:
        return None, None

    points = np.array(points)

    means = np.mean(points, axis=0)
    deviations = points - means
    # Match point set between closest points in ICP point sets.
    s_x_xp = np.sum(deviations[:, 0, 0] * deviations[:, 1, 0])
    s_y_yp = np.sum(deviations[:, 0, 1] * deviations[:, 1, 1])
    s_x_yp = np.sum(deviations[:, 0, 0] * deviations[:, 1, 1])
    s_y_xp = np.sum(deviations[:, 0, 1] * deviations[:, 1, 0])
    # Calculate rotation matrix between matched points
    rot_angle = np.arctan2(s_x_yp - s_y_xp, s_x_xp + s_y_yp)
    # Calculate translation vector between matched points
    translation = np.array([
        means[1, 0] - (means[0, 0] * np.cos(rot_angle) - means[0, 1] * np.sin(rot_angle)),
        means[1, 1] - (means[0, 0] * np.sin(rot_angle) + means[0, 1] * np.cos(rot_angle))
    ])
    # Return transformation.
    return  np.array([[math.cos(rot_angle), -math.sin(rot_angle)],
                        [math.sin(rot_angle), math.cos(rot_angle)]]), translation
    
def resample(image, output_x_size, output_y_size):
    y_size, x_size = np.shape(image)
    out_grid_x, out_grid_y = np.meshgrid(np.arange(output_x_size), np.arange(output_y_size))
    out_grid_x = out_grid_x * x_size / output_x_size
    out_grid_y = out_grid_y * y_size / output_y_size
    image = nd.map_coordinates(image, [out_grid_y, out_grid_x], order=3, cval=0.0)
    return image

def resample_both(source, target, resample_ratio):
    s_y_size, s_x_size = source.shape
    t_y_size, t_x_size = target.shape
    source = resample(source, int(s_x_size/resample_ratio), int(s_y_size/resample_ratio))
    target = resample(target, int(t_x_size/resample_ratio), int(t_y_size/resample_ratio))
    return source, target

def resample_displacement_field(u_x, u_y, output_x_size, output_y_size):
    y_size, x_size = np.shape(u_x)
    u_x = resample(u_x, output_x_size, output_y_size)
    u_y = resample(u_y, output_x_size, output_y_size)
    u_x = u_x * output_x_size/x_size
    u_y = u_y * output_y_size/y_size
    return u_x, u_y

def warp_image(image, u_x, u_y):
    y_size, x_size = image.shape
    grid_x, grid_y = np.meshgrid(np.arange(x_size), np.arange(y_size))
    return nd.map_coordinates(image, [grid_y + u_y, grid_x + u_x], order=3, cval=0.0)
def matrix_df(image, matrix):
    y_size, x_size,_ = np.shape(image)
    x_grid, y_grid = np.meshgrid(np.arange(x_size), np.arange(y_size))
    # You should use:
    points = np.vstack((x_grid.ravel(), y_grid.ravel(), np.ones_like(x_grid).ravel()))
    transformed_points = matrix @ points
    u_x = np.reshape(transformed_points[0, :], (y_size, x_size)) - x_grid
    u_y = np.reshape(transformed_points[1, :], (y_size, x_size)) - y_grid
    return u_x, u_y
    
def load_image(path):
    image = sitk.ReadImage(path)
    image = sitk.GetArrayFromImage(image)
    if image.ndim == 4:  # [z, y, x, c] format
        image = image[:, :, :3]   # Selecting the first z-slice (if needed)
    if image.ndim == 3 and image.shape[-1] >= 3:  # Ensure at least 3 channels exist
        image = image
    image = color.rgb2gray(image)
    return image

def load_landmarks(path):
    """
    Load the first two columns of landmark data from a CSV file (ignores header).
    """
    landmarks = pd.read_csv(path, header=None, usecols=[0, 1]).values.astype(np.float64)
    return landmarks


def save_landmarks(path, landmarks):
    df = pd.DataFrame(landmarks, columns=['X', 'Y'])
    df.index = np.arange(1, len(df) + 1)
    df.to_csv(path)

def pad_landmarks(landmarks, x, y):
    landmarks[:, 0] += x
    landmarks[:, 1] += y
    return landmarks

def plot_landmarks(landmarks, marker_type, colors=None):
    landmarks_length = len(landmarks)
    if colors is None:
        colors = np.random.uniform(0, 1, (3, landmarks_length))
    for i in range(landmarks_length):
        plt.plot(landmarks[i, 0], landmarks[i, 1], marker_type, color=colors[:, i])
    return colors

def normalize(image):
    return (image - np.min(image)) / (np.max(image) - np.min(image))

def to_image(array):
    return sitk.GetImageFromArray((255*array).astype(np.uint8))

def calculate_resample_size(source, target, output_max_size):
    target_y_size, target_x_size = np.shape(target)[0:2]
    source_y_size, source_x_size = np.shape(source)[0:2]

    max_y_size = max(source_y_size, target_y_size)
    max_x_size = max(source_x_size, target_x_size)

    max_dim = max(max_y_size, max_x_size)
    rescale_ratio = max_dim/output_max_size
    return rescale_ratio



def deform_conversion(displacement_field_tc: torch.Tensor):
    ndim = len(displacement_field_tc.size()) - 2
    if ndim == 2:
        displacement_field_np = displacement_field_tc.detach().clone().cpu()[0].permute(2, 0, 1).numpy()
        shape = displacement_field_np.shape
        temp_df_copy = displacement_field_np.copy()
        displacement_field_np[0, :, :] = temp_df_copy[0, :, :] / 2.0 * (shape[2])
        displacement_field_np[1, :, :] = temp_df_copy[1, :, :] / 2.0 * (shape[1])
    elif ndim == 3:
        displacement_field_np = displacement_field_tc.detach().clone().cpu()[0].permute(3, 0, 1, 2).numpy()
        shape = displacement_field_np.shape
        temp_df_copy = displacement_field_np.copy()
        displacement_field_np[0, :, :, :] = temp_df_copy[1, :, :, :] / 2.0 * (shape[2])
        displacement_field_np[1, :, :, :] = temp_df_copy[2, :, :, :] / 2.0 * (shape[1])
        displacement_field_np[2, :, :, :] = temp_df_copy[0, :, :, :] / 2.0 * (shape[3])
    return displacement_field_np

def combine_deformation(u_x, u_y, v_x, v_y):
    y_size, x_size = np.shape(u_x)
    print("u_x shape: ", u_x.shape)
    grid_x, grid_y = np.meshgrid(np.arange(x_size), np.arange(y_size))
    added_y = grid_y + v_y
    added_x = grid_x + v_x
    t_x = nd.map_coordinates(grid_x + u_x, [added_y, added_x], mode='constant', cval=0.0)
    t_y = nd.map_coordinates(grid_y + u_y, [added_y, added_x], mode='constant', cval=0.0)
    n_x, n_y = t_x - grid_x, t_y - grid_y
    indexes_x = np.logical_or(added_x >= x_size - 1, added_x <= 0)
    indexes_y = np.logical_or(added_y >= y_size - 1, added_y <= 0)
    indexes = np.logical_or(indexes_x, indexes_y)
    n_x[indexes] = 0.0
    n_y[indexes] = 0.0
    return n_x, n_y

def gaussian_filter(image, sigma):
    return nd.gaussian_filter(image, sigma)

def round_up_to_odd(value):
    return int(np.ceil(value) // 2 * 2 + 1)

def dice(image_1, image_2):
    image_1 = image_1.astype(bool)
    image_2 = image_2.astype(bool)
    return 2 * np.logical_and(image_1, image_2).sum() / (image_1.sum() + image_2.sum())
def load_image(path):
    """
    Load an image and return it as RGB (not converting to grayscale).
    """
    image = sitk.ReadImage(path)
    image = sitk.GetArrayFromImage(image)
    
    # If image is grayscale but has a singleton dimension for color, squeeze it
    if len(image.shape) == 3 and image.shape[2] == 1:
        image = np.squeeze(image)
    
    # Ensure RGB format for color images
    if len(image.shape) == 2:
        # Convert grayscale to RGB
        image = np.stack([image, image, image], axis=-1)
    elif len(image.shape) == 3 and image.shape[2] > 3:
        # Take only first 3 channels if there are more (e.g., RGBA)
        image = image[:, :, :3]
        
    return image
def load_landmarks(path):
    """
    Load landmarks from a CSV file.
    """
    landmarks = pd.read_csv(path).iloc[:, 1:].values.astype(np.float64)
    return landmarks


def transform_landmarks(landmarks, u_x, u_y):
    landmarks_x = landmarks[:, 0]
    landmarks_y = landmarks[:, 1]
    ux = nd.map_coordinates(u_x, [landmarks_y, landmarks_x], mode='nearest')
    uy = nd.map_coordinates(u_y, [landmarks_y, landmarks_x], mode='nearest')
    new_landmarks = np.stack((landmarks_x + ux, landmarks_y + uy), axis=1)
    return new_landmarks

def tre(landmarks_1, landmarks_2):
    tre = np.sqrt(np.square(landmarks_1[:, 0] - landmarks_2[:, 0]) + np.square(landmarks_1[:, 1] - landmarks_2[:, 1]))
    return tre

def rtre(landmarks_1, landmarks_2, x_size, y_size):
    return tre(landmarks_1, landmarks_2) / np.sqrt(x_size*x_size + y_size*y_size)

def print_rtre(source_landmarks, target_landmarks, x_size, y_size):
    calculated_tre = tre(source_landmarks, target_landmarks)
    mean = np.mean(calculated_tre) * 100
    median = np.median(calculated_tre) * 100
    mmax = np.max(calculated_tre) * 100
    mmin = np.min(calculated_tre) * 100
    print("TRE mean [%]: ", mean)
    print("TRE median [%]: ", median)
    print("TRE max [%]: ", mmax)
    print("TRE min [%]: ", mmin)
    return mean, median, mmax, mmin


def sort_coordinates(set1, set2):
    """
    Match points between fixed and moving point set.

    Args:
        set1 (np.array): Fixed point set
        set2 (np.array): Moving point set

    Returns:
        np.array: New moving point set
    """
    # Match points using K nearest neighbors.
    sorted_move = []
    set1 = np.array([[coord[0], coord[1]] for coord in set1])
    set2 = np.array([[coord[0], coord[1]] for coord in set2])
    knn = NearestNeighbors(n_neighbors=1, algorithm='auto').fit(set2)
    for coord in set1:
        distances, indices = knn.kneighbors([[coord[0], coord[1]]])
        sorted_move.append(set2[indices])
    # Return matched points.
    return np.array(sorted_move)

def dice(fixed, moving):
    # Calculate DICE coefficient between fixed and moving images.
    intersection = np.sum(np.logical_and(fixed, moving))
    union = np.sum(np.logical_or(fixed, moving))
    
    if union == 0:
        return 0  # handle the case where both arrays are empty
    
    dice_coefficient = 2.0 * intersection / union
    return dice_coefficient


def compute_center_of_mass(array):
    # Calculates the center of mass of point set.
    center_of_mass_2d = np.mean(array, axis=0)[:-1]
    return center_of_mass_2d

def rotate_point(x, y, cx, cy, angle):
    """
    Rotate point set around center with certain angle.

    Args:
        x (float): x-coordinate
        y (float): y-coordinate
        cx (float): center x-coordinate
        cy (float): center y-coordinate
        angle (float): rotation angle

    Returns:
        float: rotated x coordinate
        float: rotated y coordinate
    """
    # convert angles to radians
    angle_radians = math.radians(angle)
    
    # Apply rotation formula
    x_prime = (x - cx) * math.cos(angle_radians) - (y - cy) * math.sin(angle_radians) + cx
    y_prime = (x - cx) * math.sin(angle_radians) + (y - cy) * math.cos(angle_radians) + cy
    
    return x_prime, y_prime



from setuptools import setup, find_packages
import os

# Read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Read requirements
with open(os.path.join(this_directory, 'requirements.txt'), encoding='utf-8') as f:
    requirements = [line.strip() for line in f if line.strip() and not line.startswith('#')]

setup(
    name='core-registration',
    version='1.0.0',
    author='CORE Team',
    author_email='',
    description='CORE - A Cell-Level Coarse-to-Fine Image Registration Engine for Multi-stain Image Alignment',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/eshasadia/CORE',
    packages=find_packages(exclude=['tests', 'notebooks', 'docs']),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
    install_requires=requirements,
    extras_require={
        'dev': [
            'pytest>=7.0.0',
            'pytest-cov>=3.0.0',
            'black>=22.0.0',
            'isort>=5.10.0',
            'flake8>=4.0.0',
            'mypy>=0.950',
        ],
        'docs': [
            'sphinx>=4.5.0',
            'sphinx-rtd-theme>=1.0.0',
            'myst-parser>=0.17.0',
        ],
    },
    include_package_data=True,
    package_data={
        'accelerated_features': ['weights/*.pt'],
    },
    entry_points={
        'console_scripts': [
            'core-register=core.cli:main',
        ],
    },
    keywords=[
        'image registration',
        'whole slide image',
        'digital pathology',
        'multi-stain',
        'histopathology',
        'deep learning',
        'computer vision',
    ],
    project_urls={
        'Bug Reports': 'https://github.com/eshasadia/CORE/issues',
        'Source': 'https://github.com/eshasadia/CORE',
        'Documentation': 'https://core-registration.readthedocs.io/',
    },
)



import numpy as np
import cv2
import sys
import pandas as pd
import logging
from skimage import exposure
from skimage import img_as_float
from skimage.registration import phase_cross_correlation
from scipy.spatial import KDTree
from tiatoolbox.utils.metrics import dice
from tiatoolbox.tools import patchextraction
from tiatoolbox.tools.registration.wsi_registration import AffineWSITransformer
from accelerated_features.modules.xfeat import XFeat
import core.utils.util as util
from core.evaluation.evaluation import ngf_metric



# Constants
RGB_IMAGE_DIM = 3
BIN_MASK_DIM = 2

# Configure logger
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)




class Trimorph:
    @staticmethod
    def _check_dims(fixed_img, moving_img, fixed_mask, moving_mask):
        if len(np.unique(fixed_mask)) == 1 or len(np.unique(moving_mask)) == 1:
            raise ValueError("The foreground is missing in the mask.")

        if (
            fixed_img.shape[:2] != fixed_mask.shape
            or moving_img.shape[:2] != moving_mask.shape
        ):
            raise ValueError("Mismatch of shape between image and its corresponding mask.")

        if len(fixed_img.shape) == RGB_IMAGE_DIM:
            fixed_img = cv2.cvtColor(fixed_img, cv2.COLOR_BGR2GRAY)

        if len(moving_img.shape) == RGB_IMAGE_DIM:
            moving_img = cv2.cvtColor(moving_img, cv2.COLOR_BGR2GRAY)

        return fixed_img, moving_img

    @staticmethod
    def compute_center_of_mass(mask):
        moments = cv2.moments(mask)
        x_coord_center = moments["m10"] / moments["m00"]
        y_coord_center = moments["m01"] / moments["m00"]
        return (x_coord_center, y_coord_center)

    @staticmethod
    def apply_affine_transformation(fixed_img, moving_img, transform_initializer):
        return cv2.warpAffine(
            moving_img,
            transform_initializer[0:-1][:],
            fixed_img.shape[:2][::-1],
        )

    def prealignment(
        self,
        fixed_img,
        moving_img,
        fixed_mask,
        moving_mask,
        dice_overlap=0.5,
        rotation_step=10,
    ):
        orig_fixed_img, orig_moving_img = fixed_img.copy(), moving_img.copy()

        if len(fixed_mask.shape) != BIN_MASK_DIM:
            fixed_mask = fixed_mask[:, :, 0]
        if len(moving_mask.shape) != BIN_MASK_DIM:
            moving_mask = moving_mask[:, :, 0]

        fixed_mask = (fixed_mask > 0).astype(np.uint8)
        moving_mask = (moving_mask > 0).astype(np.uint8)

        fixed_img = np.squeeze(fixed_img)
        moving_img = np.squeeze(moving_img)

        fixed_img, moving_img = self._check_dims(fixed_img, moving_img, fixed_mask, moving_mask)

        if rotation_step < 10 or rotation_step > 20:
            raise ValueError("Please select the rotation step between 10 and 20.")

        if not (0 <= dice_overlap <= 1):
            raise ValueError("The dice_overlap should be in between 0 and 1.0.")

        fixed_img = exposure.rescale_intensity(img_as_float(fixed_img), in_range=(0, 1))
        moving_img = exposure.rescale_intensity(img_as_float(moving_img), in_range=(0, 1))

        height = int(np.max((fixed_mask.shape[0], moving_mask.shape[0])))
        width = int(np.max((fixed_mask.shape[1], moving_mask.shape[1])))

        padded_fixed_mask = np.pad(
            fixed_mask,
            pad_width=[(0, height - fixed_mask.shape[0]), (0, width - fixed_mask.shape[1])],
            mode="constant",
        )
        padded_moving_mask = np.pad(
            moving_mask,
            pad_width=[(0, height - moving_mask.shape[0]), (0, width - moving_mask.shape[1])],
            mode="constant",
        )
        dice_before = dice(padded_fixed_mask, padded_moving_mask)

        fixed_com = self.compute_center_of_mass((1 - fixed_img) * fixed_mask)
        moving_com = self.compute_center_of_mass((1 - moving_img) * moving_mask)

        com_transform = np.array(
            [[1, 0, fixed_com[0] - moving_com[0]],
             [0, 1, fixed_com[1] - moving_com[1]],
             [0, 0, 1]]
        )

        origin_transform_com_ = np.array([
            [1, 0, -fixed_com[0]],
            [0, 1, -fixed_com[1]],
            [0, 0, 1]
        ])
        origin_transform_com = np.array([
            [1, 0, fixed_com[0]],
            [0, 1, fixed_com[1]],
            [0, 0, 1]
        ])

        all_dice = []
        all_transform = []

        for angle in np.arange(0, 360, rotation_step).tolist():
            theta = np.radians(angle)
            c, s = np.cos(theta), np.sin(theta)
            rotation_matrix = np.array([
                [c, -s, 0],
                [s, c, 0],
                [0, 0, 1]
            ])

            transform = origin_transform_com @ rotation_matrix @ origin_transform_com_ @ com_transform

            warped_moving_mask = cv2.warpAffine(
                moving_mask,
                transform[0:-1][:],
                fixed_img.shape[:2][::-1],
            )
            dice_com = dice(fixed_mask, warped_moving_mask)

            all_dice.append(dice_com)
            all_transform.append(transform)

        if max(all_dice) >= dice_overlap:
            dice_after = max(all_dice)
            pre_transform = all_transform[all_dice.index(dice_after)]

            moving_img = self.apply_affine_transformation(orig_fixed_img, orig_moving_img, pre_transform)
            moving_mask = self.apply_affine_transformation(fixed_img, moving_mask, pre_transform)

            return pre_transform, moving_img, moving_mask, dice_after

        logger.warning(
            "Not able to find the best transformation for pre-alignment. "
            "Try changing the values for 'dice_overlap' and 'rotation_step'."
        )
        return np.eye(3), orig_moving_img, moving_mask, dice_before



"""
fixed_mask = (fixed_mask > 0).astype(np.uint8)
moving_mask = (moving_mask > 0).astype(np.uint8)

# Create instance of Trimorph
aligner = Trimorph()

# Perform prealignment
transform_matrix, aligned_image, aligned_mask, final_dice = aligner.prealignment(
    fixed_img,
    moving_img,
    fixed_mask,
    moving_mask,
    dice_overlap=0.5,
    rotation_step=10
)

# Show result (for debug)
cv2.imshow("Aligned Image", aligned_image)
cv2.imshow("Aligned Mask", aligned_mask * 255)
print("Final Dice Score:", final_dice)
cv2.waitKey(0)
cv2.destroyAllWindows()
"""
class XFeatReg:
    def __init__(self, top_k_matches: int = 8000):
        self.matcher = XFeat()
        self.top_k = top_k_matches

    def warp_corners_and_draw_matches(self, ref_points, dst_points, img1, img2):
        """
        Estimate homography, warp corners of img1 to img2, and visualize matches.
        """
        H, mask = cv2.findHomography(ref_points, dst_points, cv2.USAC_MAGSAC, 3.5, maxIters=1000, confidence=0.999)

        h, w = img1.shape[:2]
        corners_img1 = np.array([[0, 0], [w-1, 0], [w-1, h-1], [0, h-1]], dtype=np.float32).reshape(-1, 1, 2)

        if H is None:
            print("Homography computation failed. Using identity matrix.")
            H = np.eye(3, dtype=np.float32)
            mask = np.zeros((len(ref_points), 1), dtype=np.uint8)
        else:
            mask = mask.flatten()

        warped_corners = cv2.perspectiveTransform(corners_img1, H)

        img2_with_corners = img2.copy()
        for i in range(len(warped_corners)):
            start_point = tuple(warped_corners[i - 1][0].astype(int))
            end_point = tuple(warped_corners[i][0].astype(int))
            cv2.line(img2_with_corners, start_point, end_point, (0, 255, 0), 4)

        keypoints1 = [cv2.KeyPoint(p[0], p[1], 5) for p in ref_points]
        keypoints2 = [cv2.KeyPoint(p[0], p[1], 5) for p in dst_points]
        matches = [cv2.DMatch(i, i, 0) for i in range(len(mask)) if mask[i]]

        img_matches = cv2.drawMatches(
            img1, keypoints1, img2_with_corners, keypoints2, matches, None,
            matchColor=(0, 255, 0), flags=2
        )

        return img_matches

    def register(self, im1: np.ndarray, im2: np.ndarray):
        """
        Register im2 to im1 using feature matching and affine transform.
        """
        mkpts_0, mkpts_1 = self.matcher.match_xfeat_star(im2, im1, top_k=self.top_k)

        mkpts_0 = np.array(mkpts_0, dtype=np.float32)
        mkpts_1 = np.array(mkpts_1, dtype=np.float32)

        debug_matches = self.warp_corners_and_draw_matches(mkpts_0, mkpts_1, im2, im1)

        M1, mask = cv2.estimateAffinePartial2D(mkpts_0, mkpts_1, method=cv2.RANSAC)
        aligned_img = cv2.warpAffine(im2, M1, (im1.shape[1], im1.shape[0]))

        return M1, aligned_img
    
"""

# Load images
im1 = cv2.imread("/path/to/reference_image.png")
im2 = cv2.imread("/path/to/moving_image.png")

# Create aligner and register images
aligner =  XFeatReg()
M, aligned_image, match_vis = aligner.register(im1, im2)

# Display results
cv2.imshow("Aligned Image", aligned_image)
cv2.imshow("Matches", match_vis)
cv2.waitKey(0)
cv2.destroyAllWindows()

"""




def rigid_registration( moving_img, fixed_img,moving_mask, fixed_mask, verbose=False):
    """
    Aligns a moving image to a fixed image using Trimorph and optionally XFeatReg if needed.
    Args:
        fixed_img (np.ndarray): Fixed image.
        moving_img (np.ndarray): Moving image.
        fixed_mask (np.ndarray): Fixed mask.
        moving_mask (np.ndarray): Moving mask.
        verbose (bool): If True, print intermediate steps.
    Returns:
        moving_img_transformed (np.ndarray): Final aligned moving image.
        final_transform (np.ndarray): Final transformation matrix (3x3).
        timing (dict): Dictionary with timing info.
        ngf_metrics (dict): Dictionary with NGF metrics before and after optional second registration.
    """
    # Make sure you have this module
    ngf_metrics = {}
    # Initial registration with Trimorph
    aligner = Trimorph()
    best_transform1, moving_img_transformed, moving_mask_transformed, max_dice = aligner.prealignment(
        fixed_img, moving_img, fixed_mask, moving_mask
    )
    # Evaluate NGF metric
    ngf_initial = ngf_metric(fixed_img, moving_img_transformed)
    ngf_metrics['NGF trimoph'] = ngf_initial
    # If no transformation was found (identity), do extra registration
    if np.array_equal(best_transform1, np.eye(3)):
        if verbose:
            print("Running extra registration with XFeatReg...")
        aligner = XFeatReg()
        best_transform1, moving_img_transformed = aligner.register(fixed_img, moving_img)
        best_transform1 = np.vstack((best_transform1, [0, 0, 1]))
        ngf_metrics['NGF XFeat'] = ngf_metric(fixed_img, moving_img_transformed)
        final_transform = best_transform1
        return moving_img_transformed, final_transform, ngf_metrics
    else:
      # Optionally perform another registration step with XFeatReg
        aligner = XFeatReg()

        # Register using already transformed image
        M1, aligned_image = aligner.register(fixed_img, moving_img_transformed)
        ngf_second = ngf_metric(fixed_img, aligned_image)
        ngf_metrics['NGF XFeat transformed'] = ngf_second

        # Register directly from original moving image
        M2, third_aligned_image = aligner.register(fixed_img, moving_img)
        ngf_third = ngf_metric(fixed_img, third_aligned_image)
        ngf_metrics['NGF XFeat'] = ngf_third

        # Ensure M1 and M2 are 3x3 transformation matrices (homogeneous coordinates)
        M1_hom = np.vstack([M1, [0, 0, 1]]) if M1.shape == (2, 3) else M1
        M2_hom = np.vstack([M2, [0, 0, 1]]) if M2.shape == (2, 3) else M2

        # Choose the best transformation based on NGF metrics
        if ngf_initial >= ngf_second and ngf_initial >= ngf_third:
            final_transform = best_transform1
        elif ngf_second >= ngf_third:
            final_transform = best_transform1 @ M1_hom 
            moving_img_transformed = aligned_image
        else:
            final_transform = M2_hom
            moving_img_transformed = third_aligned_image

        return moving_img_transformed, final_transform, ngf_metrics

def register_wsi_with_fine_tuning(source_wsi, target_wsi, source_mask=None, target_mask=None, initial_transform=None, patch_index=10, patch_size=(5000, 5000), stride=(5000, 5000), base_resolution=0, target_resolution=10, scale_factor=0.625):
    """
    Registers a source WSI to a target WSI using an initial transform and fine-tunes the registration 
    using phase cross-correlation on a specific patch.
    
    Parameters:
    -----------
    source_wsi : object or str
        Source WSI object or path to source WSI file
    target_wsi : object or str
        Target WSI object or path to target WSI file
    source_mask : ndarray, optional
        Mask for the source WSI
    target_mask : ndarray, optional
        Mask for the target WSI
    initial_transform : ndarray, optional
        Initial homogeneous transformation matrix to apply
    patch_index : int, default=10
        Index of the patch to use for fine-tuning
    patch_size : tuple, default=(5000, 5000)
        Size of patches to extract
    stride : tuple, default=(5000, 5000)
        Stride for sliding window patch extraction
    base_resolution : int, default=0
        Base resolution for patch extraction
    target_resolution : int, default=10
        Target resolution for transformation
    scale_factor : float, default=0.625
        Scale factor to apply to the transformation matrix
        
    Returns:
    --------
    dict
        Dictionary containing:
        - 'final_transform': The final transformation matrix
        - 'registered_image': The registered image patch
        - 'fixed_tile': The fixed (target) image patch
        - 'transformed_tile': The transformed (source) image patch before fine-tuning
        - 'shift': The calculated shift between patches
    """
   
    
    # Set up patch extractors for both target and source WSIs
    fixed_patch_extractor = patchextraction.get_patch_extractor(
        input_img=target_wsi,
        method_name="slidingwindow",
        patch_size=patch_size,
        input_mask=target_mask,
        stride=stride,
        resolution=base_resolution
    )
    
    moving_patch_extractor = patchextraction.get_patch_extractor(
        input_img=source_wsi,
        method_name="slidingwindow",
        patch_size=patch_size,
        input_mask=source_mask,
        stride=stride,
        resolution=base_resolution
    )
    
    # Use the provided initial transform or identity matrix
    M_homogeneous = initial_transform if initial_transform is not None else np.eye(3)
    
    # Scale the transformation matrix for the target resolution
    transform_40x = util.scale_transformation_matrix(M_homogeneous, scale_factor, target_resolution)
    
    # Get the fixed tile from the target WSI
    fixed_tile = fixed_patch_extractor[patch_index]
    
    # Create transformer for the source WSI
    tfm = AffineWSITransformer(source_wsi, transform_40x)
    
    # Get coordinates of the selected patch
    coords = fixed_patch_extractor.coordinate_list[patch_index]
    location = (coords[0], coords[1])
    size = (coords[2] - coords[0], coords[3] - coords[1])
    
    # Read the corresponding patch from the transformed source WSI
    transformed_tile = tfm.read_rect(location, size, resolution=target_resolution, units="power")
    
    # Optional gamma correction can be uncommented if needed
    # transformed_tile = util.gamma_corrections(transformed_tile, 0.8)
    
    # Calculate shift between the fixed and transformed tiles using phase cross-correlation
    shift, error, diffphase = phase_cross_correlation(fixed_tile, transformed_tile)
    
    # Create translation matrix based on the calculated shift
    translation_offset = np.array([
        [1, 0, -shift[1]], 
        [0, 1, -shift[0]], 
        [0, 0, 1]
    ])
    
    # Apply the translation to get the registered image
    registered_image = cv2.warpAffine(
        transformed_tile, 
        translation_offset[0:-1], 
        (transformed_tile.shape[1], transformed_tile.shape[0])
    )
    
    # Update the transformation matrix with the fine-tuned translation
    transform_40x[0, 2] += translation_offset[0, 2]  # Add x translation
    transform_40x[1, 2] += translation_offset[1, 2]  # Add y translation
    
    # Recalculate the transformer with the updated transformation matrix
    updated_tfm = AffineWSITransformer(source_wsi, transform_40x)
    
    # Extract the tiles again using the updated transformer
    fixed_tile_updated = fixed_patch_extractor[patch_index]
    
    # Get the updated transformed tile using the updated transformer
    updated_transformed_tile = updated_tfm.read_rect(location, size, resolution=target_resolution, units="power")
    
    return {
        'final_transform': transform_40x,
        'registered_image': registered_image,
        'original_fixed_tile': fixed_tile,
        'original_transformed_tile': transformed_tile,
        'updated_fixed_tile': fixed_tile_updated,
        'updated_transformed_tile': updated_transformed_tile,
        'shift': shift,
        'updated_transformer': updated_tfm
    }
def extract_corresponding_tiles(source_wsi, target_wsi, source_mask=None, target_mask=None, 
                               transform_matrix=None, patch_index=10, patch_size=(5000, 5000), 
                               stride=(5000, 5000), base_resolution=0, target_resolution=10, 
                               scale_factor=0.625):
    """
    Extracts corresponding tiles from source and target WSIs using a transformation matrix.
    
    Parameters:
    -----------
    source_wsi : object or str
        Source WSI object or path to source WSI file
    target_wsi : object or str
        Target WSI object or path to target WSI file
    source_mask : ndarray, optional
        Mask for the source WSI
    target_mask : ndarray, optional
        Mask for the target WSI
    transform_matrix : ndarray, optional
        Homogeneous transformation matrix to apply
    patch_index : int, default=10
        Index of the patch to extract
    patch_size : tuple, default=(5000, 5000)
        Size of patches to extract
    stride : tuple, default=(5000, 5000)
        Stride for sliding window patch extraction
    base_resolution : int, default=0
        Base resolution for patch extraction
    target_resolution : int, default=10
        Target resolution for transformation
    scale_factor : float, default=0.625
        Scale factor to apply to the transformation matrix
        
    Returns:
    --------
    dict
        Dictionary containing:
        - 'fixed_tile': The fixed (target) image patch
        - 'transformed_tile': The transformed (source) image patch
        - 'transform_40x': The scaled transformation matrix
        - 'location': The location of the extracted patch
        - 'size': The size of the extracted patch
        - 'patch_extractor_fixed': The patch extractor for the target WSI
        - 'patch_extractor_moving': The patch extractor for the source WSI
        - 'transformer': The AffineWSITransformer used
    """
  
    
    # Set up patch extractors for both target and source WSIs
    fixed_patch_extractor = patchextraction.get_patch_extractor(
        input_img=target_wsi,
        method_name="slidingwindow",
        patch_size=patch_size,
        input_mask=target_mask,
        stride=stride,
        resolution=base_resolution
    )
    
    moving_patch_extractor = patchextraction.get_patch_extractor(
        input_img=source_wsi,
        method_name="slidingwindow",
        patch_size=patch_size,
        input_mask=source_mask,
        stride=stride,
        resolution=base_resolution
    )
    
    # Use the provided transform matrix or identity matrix
    M_homogeneous = transform_matrix if transform_matrix is not None else np.eye(3)
    
    # Scale the transformation matrix for the target resolution
    transform_40x = util.scale_transformation_matrix(M_homogeneous, scale_factor, target_resolution)
    
    # Get the fixed tile from the target WSI
    fixed_tile = fixed_patch_extractor[patch_index]
    
    # Create transformer for the source WSI
    tfm = AffineWSITransformer(source_wsi, transform_40x)
    
    # Get coordinates of the selected patch
    coords = fixed_patch_extractor.coordinate_list[patch_index]
    location = (coords[0], coords[1])
    size = (coords[2] - coords[0], coords[3] - coords[1])
    
    # Read the corresponding patch from the transformed source WSI
    transformed_tile = tfm.read_rect(location, size, resolution=target_resolution, units="power")
    
    return {
        'fixed_tile': fixed_tile,
        'transformed_tile': transformed_tile,
        'transform_40x': transform_40x,
    }


class ShapeAwarePointSetRegistration:
    def __init__(self, fixed_points, moving_points,
                 shape_attribute=None, shape_weight=0.0,
                 max_iterations=50, tolerance=1e-6,
                 allow_scaling=False):

        # --- Convert inputs to numpy arrays ---
        if isinstance(fixed_points, pd.DataFrame):
            # Support both 'x', 'y' and 'global_x', 'global_y' column names
            if 'global_x' in fixed_points.columns:
                self.fixed = fixed_points[['global_x', 'global_y']].values.astype(np.float64)
            else:
                self.fixed = fixed_points[['x', 'y']].values.astype(np.float64)
            self.fixed_shape = (fixed_points[shape_attribute].values.astype(np.float64) 
                               if shape_attribute and shape_attribute in fixed_points else None)
        else:
            self.fixed = np.asarray(fixed_points, dtype=np.float64)
            self.fixed_shape = None

        if isinstance(moving_points, pd.DataFrame):
            # Support both 'x', 'y' and 'global_x', 'global_y' column names
            if 'global_x' in moving_points.columns:
                self.moving = moving_points[['global_x', 'global_y']].values.astype(np.float64)
            else:
                self.moving = moving_points[['x', 'y']].values.astype(np.float64)
            self.moving_shape = (moving_points[shape_attribute].values.astype(np.float64)
                                if shape_attribute and shape_attribute in moving_points else None)
        else:
            self.moving = np.asarray(moving_points, dtype=np.float64)
            self.moving_shape = None

        if self.fixed.shape[1] != 2 or self.moving.shape[1] != 2:
            raise ValueError("Both fixed_points and moving_points must have exactly 2 columns (x, y).")

        self.shape_weight = np.clip(shape_weight, 0.0, 1.0)
        self.max_iterations = max_iterations
        self.tolerance = tolerance
        self.allow_scaling = allow_scaling

        # Normalize shape attributes for better weighting
        if self.fixed_shape is not None and self.moving_shape is not None:
            shape_std = np.std(self.fixed_shape)
            if shape_std > 1e-9:
                self.shape_scale = shape_std
            else:
                self.shape_scale = 1.0
        else:
            self.shape_scale = 1.0

        self.rotation = 0.0
        self.scale = 1.0
        self.translation = np.zeros(2)
        
        # Store DataFrames for visualization
        self.fixed_points = fixed_points if isinstance(fixed_points, pd.DataFrame) else None
        self.moving_points = moving_points if isinstance(moving_points, pd.DataFrame) else None
        self.registered_points = None
        self.correspondence_indices = None

    def _apply_transform(self, points, R, t, s):
        """Apply similarity transform with vectorized operations."""
        return s * np.dot(points, R.T) + t

    def _estimate_rigid_transform(self, A, B, weights=None):
        """Estimate weighted similarity transform (rotation, translation, scale)."""
        if weights is None:
            weights = np.ones(len(A))
        
        weights = weights / np.sum(weights)  # normalize weights
        
        # Weighted centroids
        centroid_A = np.sum(A * weights[:, np.newaxis], axis=0)
        centroid_B = np.sum(B * weights[:, np.newaxis], axis=0)
        
        AA = A - centroid_A
        BB = B - centroid_B
        
        # Weighted covariance
        H = np.dot(AA.T, BB * weights[:, np.newaxis])
        
        U, _, Vt = np.linalg.svd(H)
        R = np.dot(Vt.T, U.T)
        
        # Ensure proper rotation (det = 1)
        if np.linalg.det(R) < 0:
            Vt[-1, :] *= -1
            R = np.dot(Vt.T, U.T)

        if self.allow_scaling:
            num = np.sum(weights[:, np.newaxis] * BB * np.dot(AA, R))
            den = np.sum(weights[:, np.newaxis] * AA * AA)
            s = num / (den + 1e-9)
            s = max(0.1, min(s, 10.0))  # clamp scale to reasonable range
        else:
            s = 1.0

        t = centroid_B - s * np.dot(R, centroid_A)
        return R, t, s

    def register(self):
        """Perform iterative closest point registration with shape awareness."""
        fixed_xy = self.fixed
        moving_xy = self.moving
        n_moving = len(moving_xy)

        R = np.eye(2)
        t = np.zeros(2)
        s = 1.0
        prev_error = np.inf

        # Pre-compute spatial distance scale for normalization
        spatial_std = np.std(fixed_xy, axis=0).mean()
        if spatial_std < 1e-9:
            spatial_std = 1.0

        for it in range(self.max_iterations):
            transformed = self._apply_transform(moving_xy, R, t, s)

            # --- Step 1: Find spatial nearest neighbors ---
            kdtree = KDTree(fixed_xy)
            spatial_dists, spatial_idx = kdtree.query(transformed, k=1)

            # --- Step 2: Refine correspondences with shape information ---
            if self.shape_weight > 0 and self.fixed_shape is not None and self.moving_shape is not None:
                # Normalize spatial distances
                norm_spatial_dists = spatial_dists / spatial_std
                
                # Compute shape distances
                shape_dists = np.abs(self.fixed_shape[spatial_idx] - self.moving_shape) / self.shape_scale
                
                # Combined distance metric
                combined_dists = (1 - self.shape_weight) * norm_spatial_dists + self.shape_weight * shape_dists
                
                # For better matching, consider k-nearest neighbors and pick best combined match
                k_neighbors = min(5, len(fixed_xy))
                nn_dists, nn_idx = kdtree.query(transformed, k=k_neighbors)
                
                best_idx = np.zeros(n_moving, dtype=int)
                for i in range(n_moving):
                    candidates = nn_idx[i]
                    spatial_d = nn_dists[i] / spatial_std
                    shape_d = np.abs(self.fixed_shape[candidates] - self.moving_shape[i]) / self.shape_scale
                    combined = (1 - self.shape_weight) * spatial_d + self.shape_weight * shape_d
                    best_idx[i] = candidates[np.argmin(combined)]
                
                idx = best_idx
                matched_fixed = fixed_xy[idx]
                
                # Weight matches by inverse distance for robustness
                weights = 1.0 / (combined_dists + 1e-6)
            else:
                idx = spatial_idx
                matched_fixed = fixed_xy[idx]
                weights = 1.0 / (spatial_dists + 1e-6)

            # --- Step 3: Estimate transform with weighted least squares ---
            R_new, t_new, s_new = self._estimate_rigid_transform(moving_xy, matched_fixed, weights)

            # --- Step 4: Update transform ---
            R = R_new
            t = t_new
            s = s_new

            # --- Step 5: Compute alignment error ---
            transformed_new = self._apply_transform(moving_xy, R, t, s)
            errors = np.linalg.norm(matched_fixed - transformed_new, axis=1)
            error = np.mean(errors)
            
            # Check convergence
            if it > 0 and np.abs(prev_error - error) < self.tolerance * max(prev_error, 1.0):
                break
            prev_error = error

        # Final transformation
        transformed_final = self._apply_transform(moving_xy, R, t, s)

        self.rotation = np.arctan2(R[1, 0], R[0, 0])
        self.translation = t
        self.scale = s
        self.final_error = prev_error
        self.num_iterations = it + 1
        
        # Store correspondence indices for visualization
        self.correspondence_indices = idx

        self.registered_points = pd.DataFrame({
            'global_x': moving_xy[:, 0],
            'global_y': moving_xy[:, 1],
            'registered_x': transformed_final[:, 0],
            'registered_y': transformed_final[:, 1],
            'area': self.moving_shape if self.moving_shape is not None else np.ones(len(moving_xy))
        })

        return self.registered_points

    def get_transformation_matrix(self):
        """Return 3x3 homogeneous transformation matrix."""
        theta = self.rotation
        cos_t, sin_t = np.cos(theta), np.sin(theta)
        tx, ty = self.translation
        s = self.scale
        return np.array([
            [s * cos_t, -s * sin_t, tx],
            [s * sin_t,  s * cos_t, ty],
            [0,          0,         1]
        ])


def perform_shape_aware_registration(fixed_df, moving_df,
                                     shape_attribute=None,
                                     shape_weight=0.0,
                                     max_iterations=50,
                                     tolerance=1e-6,
                                     allow_scaling=False):
    """
    Convenience function for shape-aware point set registration.
    
    Returns:
        registrator: The registration object with transformation parameters
        transform_matrix: 3x3 homogeneous transformation matrix
        coords: Nx2 array of registered coordinates
    """
    registrator = ShapeAwarePointSetRegistration(
        fixed_df, moving_df,
        shape_attribute=shape_attribute,
        shape_weight=shape_weight,
        max_iterations=max_iterations,
        tolerance=tolerance,
        allow_scaling=allow_scaling
    )
    registered_points = registrator.register()
    transform_matrix = registrator.get_transformation_matrix()
    coords = registered_points[['registered_x', 'registered_y']].values
    return registrator, transform_matrix, coords


from typing import Iterable, Tuple, List, Dict
import math
import numpy as np
import numpy as np
from typing import Tuple, Dict
import numpy as np
from typing import Dict


def calculate_pad_value(size_1: Iterable[int], size_2: Iterable[int]) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
    """Calculates the padding required to make two images the same size."""
    y_size_1, x_size_1 = size_1
    y_size_2, x_size_2 = size_2
    pad_1 = [(0, 0), (0, 0)]
    pad_2 = [(0, 0), (0, 0)]
    
    if y_size_1 > y_size_2:
        pad_size = y_size_1 - y_size_2
        pad = (math.floor(pad_size / 2), math.ceil(pad_size / 2))
        pad_2[0] = pad
    elif y_size_1 < y_size_2:
        pad_size = y_size_2 - y_size_1
        pad = (math.floor(pad_size / 2), math.ceil(pad_size / 2))
        pad_1[0] = pad
        
    if x_size_1 > x_size_2:
        pad_size = x_size_1 - x_size_2
        pad = (math.floor(pad_size / 2), math.ceil(pad_size / 2))
        pad_2[1] = pad
    elif x_size_1 < x_size_2:
        pad_size = x_size_2 - x_size_1
        pad = (math.floor(pad_size / 2), math.ceil(pad_size / 2))
        pad_1[1] = pad
        
    return pad_1, pad_2

def apply_padding_landmarks(landmarks: np.ndarray, pad: List[Tuple[int, int]]) -> np.ndarray:
    """Applies padding to landmark coordinates."""
    padded_landmarks = landmarks.copy()
    padded_landmarks[:, 0] += pad[1][0]  # x coordinate
    padded_landmarks[:, 1] += pad[0][0]  # y coordinate
    return padded_landmarks



def pad_image_src(
    source: np.ndarray,
    target: np.ndarray,
    pad_value: int = 255  # white padding for uint8
) -> Tuple[np.ndarray, np.ndarray, Dict, np.ndarray, np.ndarray]:
    """
    Pads or crops the source image so it matches the target's size.
    Target remains unchanged.

    Positive padding if source < target.
    Negative padding (crop) if source > target.
    """

    src_h, src_w = source.shape[:2]
    tgt_h, tgt_w = target.shape[:2]

    # Compute differences
    dy = tgt_h - src_h
    dx = tgt_w - src_w

    pad_src = [[0, 0], [0, 0]]  # store applied padding/cropping

    # --- Y-axis (height) ---
    if dy > 0:
        # Source smaller → pad
        top = dy // 2
        bottom = dy - top
        source = np.pad(source, ((top, bottom), (0, 0), (0, 0)),
                        mode='constant', constant_values=pad_value)
        pad_src[0] = [top, bottom]
    elif dy < 0:
        # Source larger → crop
        crop = abs(dy)
        top = crop // 2
        bottom = crop - top
        source = source[top:src_h - bottom, :, :]
        pad_src[0] = [-top, -bottom]

    # --- X-axis (width) ---
    if dx > 0:
        # Source smaller → pad
        left = dx // 2
        right = dx - left
        source = np.pad(source, ((0, 0), (left, right), (0, 0)),
                        mode='constant', constant_values=pad_value)
        pad_src[1] = [left, right]
    elif dx < 0:
        # Source larger → crop
        crop = abs(dx)
        left = crop // 2
        right = crop - left
        source = source[:, left:src_w - right, :]
        pad_src[1] = [-left, -right]

    # Now both have identical shape
    assert source.shape[:2] == target.shape[:2], "Source should now match target size."

    padding_params = {
        'pad_source': pad_src
    }

    return source, target, padding_params

def pad_images(
    image_1: np.ndarray,
    image_2: np.ndarray,
    pad_value: float = 1.0
) -> Tuple[np.ndarray, np.ndarray, Dict, np.ndarray, np.ndarray]:
    """Pads two images to the same size and optionally adjusts landmarks accordingly."""
    y_size_1, x_size_1 = image_1.shape[:2]
    y_size_2, x_size_2 = image_2.shape[:2]
    
    pad_1, pad_2 = calculate_pad_value((y_size_1, x_size_1), (y_size_2, x_size_2))
    
    image_1 = np.pad(image_1, ((pad_1[0][0], pad_1[0][1]), (pad_1[1][0], pad_1[1][1]), (0, 0)), 
                     mode='constant', constant_values=pad_value)
    image_2 = np.pad(image_2, ((pad_2[0][0], pad_2[0][1]), (pad_2[1][0], pad_2[1][1]), (0, 0)), 
                     mode='constant', constant_values=pad_value)
    
  
    padding_params = {
        'pad_1': pad_1,
        'pad_2': pad_2
    }
    
    return image_1, image_2, padding_params

def pad_landmarks(
    padding_params,
    landmarks_1: np.ndarray = None,
    landmarks_2: np.ndarray = None,
    
) -> Tuple[np.ndarray, np.ndarray, Dict, np.ndarray, np.ndarray]:
   
    
    
    
    landmarks_1_padded = apply_padding_landmarks(landmarks_1, padding_params['pad_1']) if landmarks_1 is not None else None
    landmarks_2_padded = apply_padding_landmarks(landmarks_2, padding_params['pad_2']) if landmarks_2 is not None else None
    
   
    return  landmarks_1_padded, landmarks_2_padded

def remove_padding(image, pad_tuple):
    """Crop image or deformation field back to original (unpadded) size."""
    (top, bottom), (left, right) = pad_tuple
    h, w = image.shape[:2]
    cropped = image[top:h - bottom, left:w - right]
    return cropped



"""
Preprocessing module for CORE

Contains preprocessing functions for WSI images including:
- Image loading and preprocessing
- Tissue mask extraction
- Nuclei detection and analysis
- Stain normalization
- Padding utilities
"""

from core.preprocessing.preprocessing import (
    load_wsi_images,
    preprocess_images,
    extract_tissue_masks,
    adjust_gamma,
    scale_transformation_matrix,
    load_slide,
    pad_single,
    pad_images_np,
    pad_to_same_size,
)

from core.preprocessing.nuclei_analysis import (
    extract_patches_from_wsi,
    detect_nuclei_patch_watershed,
    process_fixed_patch,
    process_moving_patch,
    process_nuclei_in_patches,
    save_nuclei_data_to_csv,
    load_nuclei_coordinates,
    extract_nuclei_points,
    subsample_nuclei,
    create_nuclei_dataframe_from_points,
)

from core.preprocessing.tissuemask import (
    FlorenceTissueMaskExtractor,
    UNetTissueMaskExtractor,
)

from core.preprocessing.stainnorm import StainNormalizer

from core.preprocessing.padding import (
    calculate_pad_value,
    apply_padding_landmarks,
    pad_image_src,
    pad_images,
    pad_landmarks,
    remove_padding,
)

__all__ = [
    # From preprocessing.py
    "load_wsi_images",
    "preprocess_images",
    "extract_tissue_masks",
    "adjust_gamma",
    "scale_transformation_matrix",
    "load_slide",
    "pad_single",
    "pad_images_np",
    "pad_to_same_size",
    
    # From nuclei_analysis.py
    "extract_patches_from_wsi",
    "detect_nuclei_patch_watershed",
    "process_fixed_patch",
    "process_moving_patch",
    "process_nuclei_in_patches",
    "save_nuclei_data_to_csv",
    "load_nuclei_coordinates",
    "extract_nuclei_points",
    "subsample_nuclei",
    "create_nuclei_dataframe_from_points",
    
    # From tissuemask.py
    "FlorenceTissueMaskExtractor",
    "UNetTissueMaskExtractor",
    
    # From stainnorm.py
    "StainNormalizer",
    
    # From padding.py
    "calculate_pad_value",
    "apply_padding_landmarks",
    "pad_image_src",
    "pad_images",
    "pad_landmarks",
    "remove_padding",
]



"""
	"XFeat: Accelerated Features for Lightweight Image Matching, CVPR 2024."
	https://www.verlab.dcc.ufmg.br/descriptors/xfeat_cvpr24/

    MegaDepth data handling was adapted from 
    LoFTR official code: https://github.com/zju3dv/LoFTR/blob/master/src/datasets/megadepth.py
"""

import torch
from kornia.utils import create_meshgrid
import matplotlib.pyplot as plt
import pdb
import cv2

from modules.training.utils import plot_corrs

@torch.no_grad()
def warp_kpts(kpts0, depth0, depth1, T_0to1, K0, K1):
    """ Warp kpts0 from I0 to I1 with depth, K and Rt
    Also check covisibility and depth consistency.
    Depth is consistent if relative error < 0.2 (hard-coded).
    
    Args:
        kpts0 (torch.Tensor): [N, L, 2] - <x, y>,
        depth0 (torch.Tensor): [N, H, W],
        depth1 (torch.Tensor): [N, H, W],
        T_0to1 (torch.Tensor): [N, 3, 4],
        K0 (torch.Tensor): [N, 3, 3],
        K1 (torch.Tensor): [N, 3, 3],
    Returns:
        calculable_mask (torch.Tensor): [N, L]
        warped_keypoints0 (torch.Tensor): [N, L, 2] <x0_hat, y1_hat>
    """
    kpts0_long = kpts0.round().long().clip(0, 2000-1)

    depth0[:, 0, :] = 0 ; depth1[:, 0, :] = 0 
    depth0[:, :, 0] = 0 ; depth1[:, :, 0] = 0 

    # Sample depth, get calculable_mask on depth != 0
    kpts0_depth = torch.stack(
        [depth0[i, kpts0_long[i, :, 1], kpts0_long[i, :, 0]] for i in range(kpts0.shape[0])], dim=0
    )  # (N, L)
    nonzero_mask = kpts0_depth > 0

    # Draw cross marks on the image for each keypoint
    # for b in range(len(kpts0)):
    #     fig, ax = plt.subplots(1,2)
    #     depth_np = depth0.numpy()[b]
    #     depth_np_plot = depth_np.copy()
    #     for x, y in kpts0_long[b, nonzero_mask[b], :].numpy():
    #         cv2.drawMarker(depth_np_plot, (x, y), (255), cv2.MARKER_CROSS, markerSize=10, thickness=2)
    #     ax[0].imshow(depth_np)
    #     ax[1].imshow(depth_np_plot)

    # Unproject
    kpts0_h = torch.cat([kpts0, torch.ones_like(kpts0[:, :, [0]])], dim=-1) * kpts0_depth[..., None]  # (N, L, 3)
    kpts0_cam = K0.inverse() @ kpts0_h.transpose(2, 1)  # (N, 3, L)

    # Rigid Transform
    w_kpts0_cam = T_0to1[:, :3, :3] @ kpts0_cam + T_0to1[:, :3, [3]]    # (N, 3, L)
    w_kpts0_depth_computed = w_kpts0_cam[:, 2, :]

    # Project
    w_kpts0_h = (K1 @ w_kpts0_cam).transpose(2, 1)  # (N, L, 3)
    w_kpts0 = w_kpts0_h[:, :, :2] / (w_kpts0_h[:, :, [2]] + 1e-5)  # (N, L, 2), +1e-4 to avoid zero depth

    # Covisible Check
    # h, w = depth1.shape[1:3]
    # covisible_mask = (w_kpts0[:, :, 0] > 0) * (w_kpts0[:, :, 0] < w-1) * \
    #     (w_kpts0[:, :, 1] > 0) * (w_kpts0[:, :, 1] < h-1)
    # w_kpts0_long = w_kpts0.long()
    # w_kpts0_long[~covisible_mask, :] = 0

    # w_kpts0_depth = torch.stack(
    #     [depth1[i, w_kpts0_long[i, :, 1], w_kpts0_long[i, :, 0]] for i in range(w_kpts0_long.shape[0])], dim=0
    # )  # (N, L)
    # consistent_mask = ((w_kpts0_depth - w_kpts0_depth_computed) / w_kpts0_depth).abs() < 0.2


    valid_mask = nonzero_mask #* consistent_mask* covisible_mask 

    return valid_mask, w_kpts0


@torch.no_grad()
def spvs_coarse(data, scale = 8):
    """
        Supervise corresp with dense depth & camera poses
    """

    # 1. misc
    device = data['image0'].device
    N, _, H0, W0 = data['image0'].shape
    _, _, H1, W1 = data['image1'].shape
    #scale = 8
    scale0 = scale * data['scale0'][:, None] if 'scale0' in data else scale
    scale1 = scale * data['scale1'][:, None] if 'scale1' in data else scale
    h0, w0, h1, w1 = map(lambda x: x // scale, [H0, W0, H1, W1])

    # 2. warp grids
    # create kpts in meshgrid and resize them to image resolution
    grid_pt1_c = create_meshgrid(h1, w1, False, device).reshape(1, h1*w1, 2).repeat(N, 1, 1)    # [N, hw, 2]
    grid_pt1_i = scale1 * grid_pt1_c

    # warp kpts bi-directionally and check reproj error
    nonzero_m1, w_pt1_i  =  warp_kpts(grid_pt1_i, data['depth1'], data['depth0'], data['T_1to0'], data['K1'], data['K0']) 
    nonzero_m2, w_pt1_og =  warp_kpts(   w_pt1_i, data['depth0'], data['depth1'], data['T_0to1'], data['K0'], data['K1']) 


    dist = torch.linalg.norm( grid_pt1_i - w_pt1_og, dim=-1)
    mask_mutual = (dist < 1.5) & nonzero_m1 & nonzero_m2

    #_, w_pt1_i = warp_kpts(grid_pt1_i, data['depth1'], data['depth0'], data['T_1to0'], data['K1'], data['K0'])
    batched_corrs = [ torch.cat([w_pt1_i[i, mask_mutual[i]] / data['scale0'][i],
                       grid_pt1_i[i, mask_mutual[i]] / data['scale1'][i]],dim=-1) for i in range(len(mask_mutual))]


    #Remove repeated correspondences - this is important for network convergence
    corrs = []
    for pts in batched_corrs:
        lut_mat12 = torch.ones((h1, w1, 4), device = device, dtype = torch.float32) * -1
        lut_mat21 = torch.clone(lut_mat12)
        src_pts = pts[:, :2] / scale
        tgt_pts = pts[:, 2:] / scale
        try:
            lut_mat12[src_pts[:,1].long(), src_pts[:,0].long()] = torch.cat([src_pts, tgt_pts], dim=1)
            mask_valid12 = torch.all(lut_mat12 >= 0, dim=-1)
            points = lut_mat12[mask_valid12]

            #Target-src check
            src_pts, tgt_pts = points[:, :2], points[:, 2:]
            lut_mat21[tgt_pts[:,1].long(), tgt_pts[:,0].long()] = torch.cat([src_pts, tgt_pts], dim=1)
            mask_valid21 = torch.all(lut_mat21 >= 0, dim=-1)
            points = lut_mat21[mask_valid21]

            corrs.append(points)
        except:
            pdb.set_trace()
            print('..')

    #Plot for debug purposes    
    # for i in range(len(corrs)):
    #     plot_corrs(data['image0'][i], data['image1'][i], corrs[i][:, :2]*8, corrs[i][:, 2:]*8)

    return corrs

@torch.no_grad()
def get_correspondences(pts2, data, idx):
    device = data['image0'].device
    N, _, H0, W0 = data['image0'].shape
    _, _, H1, W1 = data['image1'].shape

    pts2 = pts2[None, ...]

    scale0 = data['scale0'][idx, None][None, ...] if 'scale0' in data else 1
    scale1 = data['scale1'][idx, None][None, ...] if 'scale1' in data else 1

    pts2 = scale1 * pts2 * 8

    # warp kpts bi-directionally and check reproj error
    nonzero_m1, pts1  = warp_kpts(pts2, data['depth1'][idx][None, ...], data['depth0'][idx][None, ...], data['T_1to0'][idx][None, ...], 
                                                                        data['K1'][idx][None, ...], data['K0'][idx][None, ...]) 

    corrs = torch.cat([pts1[0, :] / data['scale0'][idx],
                       pts2[0, :] / data['scale1'][idx]],dim=-1)

    #plot_corrs(data['image0'][idx], data['image1'][idx], corrs[:, :2], corrs[:, 2:])

    return corrs




"""
MHA Deformation Field Application for CORE

Apply deformation fields computed at low resolution (e.g., 0.625x) 
to whole slide images at high resolution (e.g., 40x).

Features:
- Uses pyvips for memory-efficient large image handling
- Scales deformation fields from low to high resolution
- Properly scales both spatial dimensions and displacement magnitudes
"""

import numpy as np
import pyvips
import SimpleITK as sitk
from pathlib import Path
from typing import Tuple, Optional, Union
import argparse


def get_image_shape(img: Union[np.ndarray, pyvips.Image]) -> Tuple[int, ...]:
    """Get shape of image (row, col) or (row, col, channels)."""
    if isinstance(img, pyvips.Image):
        if img.bands > 1:
            return (img.height, img.width, img.bands)
        return (img.height, img.width)
    return img.shape


def array_to_vips(arr: np.ndarray) -> pyvips.Image:
    """Convert numpy array to pyvips Image."""
    if arr.ndim == 2:
        height, width = arr.shape
        bands = 1
        linear = arr.flatten()
    else:
        height, width, bands = arr.shape
        linear = arr.flatten()
    
    dtype_to_format = {
        np.dtype('uint8'): 'uchar',
        np.dtype('int8'): 'char',
        np.dtype('uint16'): 'ushort',
        np.dtype('int16'): 'short',
        np.dtype('uint32'): 'uint',
        np.dtype('int32'): 'int',
        np.dtype('float32'): 'float',
        np.dtype('float64'): 'double',
    }
    
    vips_format = dtype_to_format.get(arr.dtype, 'float')
    
    vips_img = pyvips.Image.new_from_memory(
        linear.tobytes(),
        width, height, bands,
        vips_format
    )
    
    return vips_img


def vips_to_array(vips_img: pyvips.Image) -> np.ndarray:
    """Convert pyvips Image to numpy array."""
    format_map = {
        'uchar': np.uint8,
        'char': np.int8,
        'ushort': np.uint16,
        'short': np.int16,
        'uint': np.uint32,
        'int': np.int32,
        'float': np.float32,
        'double': np.float64,
    }
    
    dtype = np.dtype(format_map.get(vips_img.format, np.float32))
    
    np_arr = np.ndarray(
        buffer=vips_img.write_to_memory(),
        dtype=dtype,
        shape=[vips_img.height, vips_img.width, vips_img.bands]
    )
    
    if np_arr.shape[2] == 1:
        np_arr = np_arr.squeeze(axis=2)
    
    return np_arr


def resize_displacement_field(
    displacement: Union[np.ndarray, Tuple[np.ndarray, np.ndarray]],
    target_shape_rc: Tuple[int, int],
    source_shape_rc: Tuple[int, int],
    registered_shape_rc: Tuple[int, int]
) -> np.ndarray:
    """
    Resize displacement field to match target resolution.
    
    Steps:
    1. Resize the displacement field spatially to match target dimensions
    2. Scale the displacement values by the ratio of resolutions
    
    Args:
        displacement: Displacement field as (2, H, W) array or tuple of (dx, dy)
        target_shape_rc: Target output shape (rows, cols)
        source_shape_rc: Shape of image where transform was computed
        registered_shape_rc: Shape of registered image at transform resolution
    
    Returns:
        Resized displacement field as (2, H, W) array
    """
    # Convert tuple to array if needed
    if isinstance(displacement, tuple):
        displacement = np.array([displacement[0], displacement[1]])
    
    # Calculate scaling factors
    scale_factor_rc = np.array(target_shape_rc) / np.array(registered_shape_rc)
    
    # Extract x and y components
    disp_x = displacement[0]
    disp_y = displacement[1]
    
    # Use pyvips for efficient resizing
    vips_disp_x = array_to_vips(disp_x.astype(np.float32))
    vips_disp_y = array_to_vips(disp_y.astype(np.float32))
    
    # Resize to target dimensions
    scale_x = target_shape_rc[1] / disp_x.shape[1]
    scale_y = target_shape_rc[0] / disp_x.shape[0]
    
    resized_disp_x = vips_disp_x.resize(scale_x, vscale=scale_y)
    resized_disp_y = vips_disp_y.resize(scale_x, vscale=scale_y)
    
    # Convert back to numpy
    scaled_disp_x = vips_to_array(resized_disp_x)
    scaled_disp_y = vips_to_array(resized_disp_y)
    
    # Scale the displacement values
    scaled_disp_x = scaled_disp_x * scale_factor_rc[1]  # x by column scale
    scaled_disp_y = scaled_disp_y * scale_factor_rc[0]  # y by row scale
    
    return np.array([scaled_disp_x, scaled_disp_y])


def apply_displacement_field(
    img: Union[np.ndarray, pyvips.Image],
    displacement: np.ndarray,
    output_shape_rc: Tuple[int, int],
    source_shape_rc: Tuple[int, int],
    registered_shape_rc: Tuple[int, int],
    background: Optional[Union[int, float, list]] = None,
    interpolation: str = "bicubic"
) -> pyvips.Image:
    """
    Apply displacement field to warp an image.
    
    Args:
        img: Input image to warp
        displacement: Backwards displacement field (2, H, W)
        output_shape_rc: Output shape (rows, cols)
        source_shape_rc: Shape where transform was computed
        registered_shape_rc: Shape of registered image at transform resolution
        background: Background color for areas outside the image
        interpolation: Interpolation method
    
    Returns:
        Warped pyvips Image
    """
    # Convert numpy to vips if needed
    if isinstance(img, np.ndarray):
        vips_img = array_to_vips(img)
    else:
        vips_img = img
    
    # Get image properties
    n_channels = vips_img.bands
    
    # Set default background
    if background is None:
        background = [0] * n_channels
    elif isinstance(background, (int, float)):
        background = [background] * n_channels
    
    # Resize displacement field to output dimensions
    scaled_displacement = resize_displacement_field(
        displacement,
        output_shape_rc,
        source_shape_rc,
        registered_shape_rc
    )
    
    # Create coordinate grids for mapping
    rows, cols = output_shape_rc
    
    col_indices = np.tile(np.arange(cols, dtype=np.float32), (rows, 1))
    row_indices = np.tile(np.arange(rows, dtype=np.float32).reshape(-1, 1), (1, cols))
    
    # Add displacements to get source coordinates
    source_cols = col_indices + scaled_displacement[0]
    source_rows = row_indices + scaled_displacement[1]
    
    # Stack into index image for mapim (2 bands: x, y)
    coord_map = np.stack([source_cols, source_rows], axis=2).astype(np.float32)
    vips_coord_map = array_to_vips(coord_map)
    
    # Set interpolation method
    interp_methods = {
        'nearest': pyvips.Interpolate.new('nearest'),
        'bilinear': pyvips.Interpolate.new('bilinear'),
        'bicubic': pyvips.Interpolate.new('bicubic'),
    }
    interpolator = interp_methods.get(interpolation, pyvips.Interpolate.new('bicubic'))
    
    # Warp image using coordinate mapping
    warped = vips_img.mapim(vips_coord_map, interpolate=interpolator, background=background)
    
    return warped


def load_mha_displacement(mha_path: str) -> Tuple[np.ndarray, dict]:
    """
    Load MHA deformation field and convert to displacement array format.
    
    Output format is (2, H, W) where:
    - displacement[0] = dx (column/x displacement)
    - displacement[1] = dy (row/y displacement)
    
    Args:
        mha_path: Path to MHA file
    
    Returns:
        Tuple of (displacement array, metadata dict)
    """
    print(f"Loading deformation field: {mha_path}")
    
    sitk_img = sitk.ReadImage(mha_path)
    
    # Get metadata
    metadata = {
        'size': sitk_img.GetSize(),
        'spacing': sitk_img.GetSpacing(),
        'origin': sitk_img.GetOrigin(),
        'direction': sitk_img.GetDirection(),
    }
    
    print(f"  Size: {metadata['size']}")
    print(f"  Spacing: {metadata['spacing']}")
    
    # Convert to numpy
    arr = sitk.GetArrayFromImage(sitk_img)
    
    # Handle different array shapes
    if arr.ndim == 3:
        if arr.shape[2] == 2:
            # (H, W, 2) -> (2, H, W)
            disp_x = arr[:, :, 0]
            disp_y = arr[:, :, 1]
        elif arr.shape[0] == 2:
            # Already (2, H, W)
            disp_x = arr[0]
            disp_y = arr[1]
        else:
            raise ValueError(f"Unexpected deformation field shape: {arr.shape}")
    else:
        raise ValueError(f"Expected 3D array, got shape: {arr.shape}")
    
    displacement = np.array([disp_x, disp_y], dtype=np.float32)
    print(f"  Displacement shape: {displacement.shape}")
    
    return displacement, metadata


def register_wsi_with_mha(
    mha_path: str,
    wsi_path: str,
    output_path: str,
    source_magnification: float = 0.625,
    target_magnification: float = 40.0,
    interpolation: str = "bicubic",
    background: int = 0
):
    """
    Apply MHA deformation field to WSI for registration.
    
    Args:
        mha_path: Path to deformation field MHA file
        wsi_path: Path to WSI to be warped
        output_path: Output path for registered image
        source_magnification: Magnification at which deformation was computed
        target_magnification: Target WSI magnification
        interpolation: Interpolation method
        background: Background color
    """
    print("=" * 60)
    print("WSI Registration with MHA Deformation Field")
    print("=" * 60)
    
    # Calculate scale factor
    scale_factor = target_magnification / source_magnification
    print(f"Scale factor: {scale_factor}x ({source_magnification}x -> {target_magnification}x)")
    print()
    
    # Load deformation field
    displacement, mha_metadata = load_mha_displacement(mha_path)
    source_shape_rc = (displacement.shape[1], displacement.shape[2])
    registered_shape_rc = source_shape_rc
    print()
    
    # Load WSI with pyvips for memory efficiency
    print(f"Loading WSI: {wsi_path}")
    vips_img = pyvips.Image.new_from_file(wsi_path, access='sequential')
    print(f"  Size: {vips_img.width} x {vips_img.height}")
    print(f"  Bands: {vips_img.bands}")
    print()
    
    # Calculate output shape
    output_shape_rc = (vips_img.height, vips_img.width)
    
    # Apply deformation
    print("Applying deformation field...")
    warped = apply_displacement_field(
        img=vips_img,
        displacement=displacement,
        output_shape_rc=output_shape_rc,
        source_shape_rc=source_shape_rc,
        registered_shape_rc=registered_shape_rc,
        background=background,
        interpolation=interpolation
    )
    print("  Warping complete")
    print()
    
    # Save result
    print(f"Saving to: {output_path}")
    
    # Ensure output directory exists
    Path(output_path).parent.mkdir(parents=True, exist_ok=True)
    
    # Determine output format based on extension
    ext = Path(output_path).suffix.lower()
    
    if ext == '.mha':
        # Convert to numpy and save with SimpleITK
        warped_arr = vips_to_array(warped)
        sitk_img = sitk.GetImageFromArray(warped_arr)
        sitk.WriteImage(sitk_img, output_path, useCompression=True)
    elif ext in ['.tiff', '.tif']:
        # Save as TIFF with pyvips
        warped.write_to_file(output_path, compression='lzw')
    elif ext == '.ome.tiff' or output_path.endswith('.ome.tiff'):
        # Save as OME-TIFF (pyramidal)
        warped.write_to_file(
            output_path,
            tile=True,
            pyramid=True,
            compression='lzw',
            tile_width=256,
            tile_height=256
        )
    else:
        # Default: save with pyvips
        warped.write_to_file(output_path)
    
    print("  Save complete")
    print()
    print("=" * 60)
    print("Registration complete!")
    print("=" * 60)


def save_displacement_as_mha(
    displacement: np.ndarray,
    output_path: str,
    spacing: Tuple[float, float] = (1.0, 1.0),
    origin: Tuple[float, float] = (0.0, 0.0)
):
    """
    Save displacement array as MHA file.
    
    Args:
        displacement: Displacement field (2, H, W)
        output_path: Output path
        spacing: Pixel spacing
        origin: Image origin
    """
    # Convert (2, H, W) to (H, W, 2) for SimpleITK
    disp_x = displacement[0]
    disp_y = displacement[1]
    arr = np.stack([disp_x, disp_y], axis=2).astype(np.float64)
    
    # Create SimpleITK image
    sitk_img = sitk.GetImageFromArray(arr, isVector=True)
    sitk_img.SetSpacing(spacing)
    sitk_img.SetOrigin(origin)
    
    # Save
    sitk.WriteImage(sitk_img, output_path, useCompression=True)
    print(f"Saved displacement field to: {output_path}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Apply MHA deformation field to WSI for registration"
    )
    parser.add_argument(
        "--mha", "-m",
        required=True,
        help="Path to MHA deformation field"
    )
    parser.add_argument(
        "--wsi", "-w",
        required=True,
        help="Path to WSI to be registered"
    )
    parser.add_argument(
        "--output", "-o",
        required=True,
        help="Output path"
    )
    parser.add_argument(
        "--source-mag",
        type=float,
        default=0.625,
        help="Source magnification (default: 0.625)"
    )
    parser.add_argument(
        "--target-mag",
        type=float,
        default=40.0,
        help="Target magnification (default: 40.0)"
    )
    parser.add_argument(
        "--interp",
        default="bicubic",
        choices=["nearest", "bilinear", "bicubic"],
        help="Interpolation method"
    )
    
    args = parser.parse_args()
    
    register_wsi_with_mha(
        mha_path=args.mha,
        wsi_path=args.wsi,
        output_path=args.output,
        source_magnification=args.source_mag,
        target_magnification=args.target_mag,
        interpolation=args.interp
    )
